from django.test import TransactionTestCase
from reversion.models import Version

from waldur_core.core.utils import silent_call

from . import models as test_models


class TestQuotaField(TransactionTestCase):

    def test_quota_is_automatically_created_with_scope(self):
        scope = test_models.GrandparentModel.objects.create()
        self.assertTrue(scope.quotas.filter(name=test_models.GrandparentModel.Quotas.regular_quota).exists())

    def test_quota_limit_field_create(self):
        child = test_models.GrandparentModel.objects.create(regular_quota=7)
        self.assertEqual(child.quotas.get(name='regular_quota').limit, 7)

    def test_quota_limit_field_update(self):
        child = test_models.GrandparentModel.objects.create()
        child.regular_quota = 9
        child.save()
        self.assertEqual(child.quotas.get(name='regular_quota').limit, 9)

    # XXX: Ideally this method should belong to ReversionMixin tests and
    #      should be separated into several smaller tests.
    def test_quota_versions(self):
        scope = test_models.GrandparentModel.objects.create()
        quota = scope.quotas.get(name=test_models.GrandparentModel.Quotas.regular_quota)
        quota.usage = 13.0
        quota.save()
        # make sure that new version was created after quota usage change.
        latest_version = Version.objects.get_for_object(quota).latest('revision__date_created')
        self.assertEqual(latest_version._object_version.object.usage, quota.usage)
        # make sure that new version was not created if object was saved without data change.
        quota.usage = 13
        quota.save()
        new_latest_version = Version.objects.get_for_object(quota).latest('revision__date_created')
        self.assertEqual(new_latest_version, latest_version)


class TestCounterQuotaField(TransactionTestCase):

    def setUp(self):
        self.grandparent = test_models.GrandparentModel.objects.create()
        self.parent = test_models.ParentModel.objects.create(parent=self.grandparent)
        self.child = test_models.ChildModel.objects.create(parent=self.parent)
        self.quota_field = test_models.ParentModel.Quotas.counter_quota

    def test_counter_quota_usage_is_increased_on_child_creation(self):
        quota = self.parent.quotas.get(name=self.quota_field)
        self.assertEqual(quota.usage, 1)

    def test_counter_quota_usage_is_decreased_on_child_deletion(self):
        self.child.delete()
        quota = self.parent.quotas.get(name=self.quota_field)
        self.assertEqual(quota.usage, 0)

    def test_counter_quota_usage_is_right_after_recalculation(self):
        quota = self.parent.quotas.get(name=self.quota_field)
        quota.usage = 3
        quota.save()

        silent_call('recalculatequotas')

        quota = self.parent.quotas.get(name=self.quota_field)
        self.assertEqual(quota.usage, 1)

    def test_counter_quota_usage_is_working_with_two_models_as_targets(self):
        self.parent.second_children.create()

        quota = self.parent.quotas.get(name=test_models.ParentModel.Quotas.two_targets_counter_quota)
        self.assertEqual(quota.usage, 2)

    def test_delta_quota_usage_is_increased_on_child_creation(self):
        quota = self.parent.quotas.get(name=test_models.ParentModel.Quotas.delta_quota)
        self.assertEqual(quota.usage, 10)

    def test_delta_quota_usage_is_decreased_on_child_deletion(self):
        self.child.delete()
        quota = self.parent.quotas.get(name=test_models.ParentModel.Quotas.delta_quota)
        self.assertEqual(quota.usage, 0)


class TestTotalQuotaField(TransactionTestCase):

    def setUp(self):
        self.grandparent = test_models.GrandparentModel.objects.create()
        self.parent = test_models.ParentModel.objects.create(parent=self.grandparent)
        self.child = test_models.SecondChildModel.objects.create(parent=self.parent, size=100)
        self.quota_field = test_models.ParentModel.Quotas.total_quota

    def test_counter_quota_usage_is_increased_on_child_creation(self):
        quota = self.parent.quotas.get(name=self.quota_field)
        self.assertEqual(quota.usage, 100)

    def test_counter_quota_usage_is_decreased_on_child_deletion(self):
        self.child.delete()
        quota = self.parent.quotas.get(name=self.quota_field)
        self.assertEqual(quota.usage, 0)

    def test_counter_quota_usage_is_right_after_recalculation(self):
        quota = self.parent.quotas.get(name=self.quota_field)
        quota.usage = 0
        quota.save()

        silent_call('recalculatequotas')

        quota = self.parent.quotas.get(name=self.quota_field)
        self.assertEqual(quota.usage, 100)


class TestUsageAggregatorField(TransactionTestCase):

    def setUp(self):
        self.grandparent = test_models.GrandparentModel.objects.create()
        self.parents = [test_models.ParentModel.objects.create(parent=self.grandparent) for _ in range(2)]
        non_quota_parent = test_models.NonQuotaParentModel.objects.create()
        self.children = [test_models.ChildModel.objects.create(parent=parent, non_quota_parent=non_quota_parent)
                         for parent in self.parents]

        self.child_quota_field = test_models.ChildModel.Quotas.usage_aggregator_quota
        self.parent_quota_field = test_models.ParentModel.Quotas.usage_aggregator_quota
        self.grandparent_quota_field = test_models.GrandparentModel.Quotas.usage_aggregator_quota

    def test_aggregator_usage_increases_on_child_quota_usage_increase(self):
        usage_value = 10
        for child in self.children:
            quota = child.quotas.get(name=self.child_quota_field)
            quota.usage = usage_value
            quota.save()

        for parent in self.parents:
            quota = parent.quotas.get(name=self.parent_quota_field)
            self.assertEqual(quota.usage, usage_value)

        quota = self.grandparent.quotas.get(name=self.grandparent_quota_field)
        self.assertEqual(quota.usage, usage_value * len(self.children))

    def test_aggregator_usage_decreases_on_child_deletion(self):
        usage_value = 10
        for child in self.children:
            quota = child.quotas.get(name=self.child_quota_field)
            quota.usage = usage_value
            quota.save()

        first_child = self.children[0]
        first_child.delete()
        quota = first_child.parent.quotas.get(name=self.parent_quota_field)
        self.assertEqual(quota.usage, 0)

        quota = self.grandparent.quotas.get(name=self.grandparent_quota_field)
        self.assertEqual(quota.usage, usage_value)

    def test_aggregator_usage_increases_on_child_creation(self):
        usage_value = 10
        test_models.ChildModel.Quotas.usage_aggregator_quota.default_usage = usage_value

        parent = self.parents[0]
        test_models.ChildModel.objects.create(parent=parent)

        quota = parent.quotas.get(name=self.parent_quota_field)
        self.assertEqual(quota.usage, usage_value)
        quota = self.grandparent.quotas.get(name=self.parent_quota_field)
        self.assertEqual(quota.usage, usage_value)

    def test_usage_aggregator_recalculation(self):
        usage_value = 10
        for child in self.children:
            quota = child.quotas.get(name=self.child_quota_field)
            quota.usage = usage_value
            quota.save()
        # set quota as wrong number to test recalculation
        for parent in self.parents:
            parent.set_quota_usage(self.parent_quota_field, 666)
        self.grandparent.set_quota_usage(self.grandparent_quota_field, 1232)

        silent_call('recalculatequotas')

        for parent in self.parents:
            quota = parent.quotas.get(name=self.parent_quota_field)
            self.assertEqual(quota.usage, usage_value)

        quota = self.grandparent.quotas.get(name=self.grandparent_quota_field)
        self.assertEqual(quota.usage, usage_value * len(self.children))

    def test_usage_aggregator_quota_works_with_specified_child_quota_name(self):
        usage_value = 10
        for child in self.children:
            quota = child.quotas.get(name=self.child_quota_field)
            quota.usage = usage_value
            quota.save()

        # second_usage_aggregator_quota quota should increases too
        for parent in self.parents:
            quota = parent.quotas.get(name=test_models.ParentModel.Quotas.second_usage_aggregator_quota)
            self.assertEqual(quota.usage, usage_value)


class TestLimitAggregatorField(TransactionTestCase):

    def setUp(self):
        self.grandparent = test_models.GrandparentModel.objects.create()
        self.parents = [test_models.ParentModel.objects.create(parent=self.grandparent) for _ in range(2)]
        non_quota_parent = test_models.NonQuotaParentModel.objects.create()
        self.children = [test_models.ChildModel.objects.create(parent=parent, non_quota_parent=non_quota_parent)
                         for parent in self.parents]

        self.child_quota_field = test_models.ChildModel.Quotas.limit_aggregator_quota
        self.parent_quota_field = test_models.ParentModel.Quotas.limit_aggregator_quota
        self.grandparent_quota_field = test_models.GrandparentModel.Quotas.limit_aggregator_quota

    def test_aggregator_usage_increases_on_child_quota_limit_increase(self):
        limit_value = 10
        for child in self.children:
            quota = child.quotas.get(name=self.child_quota_field)
            quota.limit = limit_value
            quota.save()

        for parent in self.parents:
            quota = parent.quotas.get(name=self.parent_quota_field)
            self.assertEqual(quota.usage, limit_value)

        quota = self.grandparent.quotas.get(name=self.grandparent_quota_field)
        self.assertEqual(quota.usage, limit_value * len(self.children))

    def test_aggregator_usage_decreases_on_child_deletion(self):
        limit_value = 10
        for child in self.children:
            quota = child.quotas.get(name=self.child_quota_field)
            quota.limit = limit_value
            quota.save()

        first_child = self.children[0]
        first_child.delete()
        quota = first_child.parent.quotas.get(name=self.parent_quota_field)
        self.assertEqual(quota.usage, 0)

        quota = self.grandparent.quotas.get(name=self.grandparent_quota_field)
        self.assertEqual(quota.usage, limit_value)

    def test_aggregator_usage_increases_on_child_creation(self):
        limit_value = 10
        test_models.ChildModel.Quotas.limit_aggregator_quota.default_limit = limit_value

        parent = self.parents[0]
        test_models.ChildModel.objects.create(parent=parent)

        quota = parent.quotas.get(name=self.parent_quota_field)
        self.assertEqual(quota.usage, limit_value)
        quota = self.grandparent.quotas.get(name=self.grandparent_quota_field)
        self.assertEqual(quota.usage, limit_value)

    def test_limit_aggregator_recalculation(self):
        limit_value = 10
        for child in self.children:
            quota = child.quotas.get(name=self.child_quota_field)
            quota.limit = limit_value
            quota.save()
        # set quota as wrong number to test recalculation
        for parent in self.parents:
            parent.set_quota_limit(self.parent_quota_field, 666)
        self.grandparent.set_quota_limit(self.grandparent_quota_field, 1232)

        silent_call('recalculatequotas')

        for parent in self.parents:
            quota = parent.quotas.get(name=self.parent_quota_field)
            self.assertEqual(quota.usage, limit_value)

        quota = self.grandparent.quotas.get(name=self.grandparent_quota_field)
        self.assertEqual(quota.usage, limit_value * len(self.children))



from django.contrib.admin.sites import AdminSite
from django.test import TestCase

from waldur_core.core.admin import UserChangeForm, UserAdmin
from waldur_core.core.models import User
from waldur_core.core.tests.helpers import override_waldur_core_settings
from waldur_core.structure.admin import CustomerAdmin
from waldur_core.structure.models import Customer
from waldur_core.structure.tests.factories import UserFactory, CustomerFactory


class MockRequest:
    pass


class MockSuperUser:
    def has_perm(self, perm):
        return True


request = MockRequest()
request.user = MockSuperUser()


class UserAdminTest(TestCase):
    def change_user(self, **kwargs):
        user = UserFactory()
        form_for_data = UserChangeForm(instance=user)

        post_data = form_for_data.initial
        post_data.update(kwargs)

        form = UserChangeForm(instance=user, data=post_data)
        form.save()

        user.refresh_from_db()
        return user

    def test_civil_number_is_stripped(self):
        user = self.change_user(civil_number='  NEW_CIVIL_NUMBER  ')
        self.assertEqual(user.civil_number, 'NEW_CIVIL_NUMBER')

    def test_whitspace_civil_number_converts_to_none(self):
        user = self.change_user(civil_number='  ')
        self.assertEqual(user.civil_number, None)

    def test_empty_civil_number_converts_to_none(self):
        user = self.change_user(civil_number='')
        self.assertEqual(user.civil_number, None)


class NativeNameAdminTest(TestCase):

    @override_waldur_core_settings(NATIVE_NAME_ENABLED=False)
    def test_native_name_is_omitted_in_user_admin_if_feature_is_not_enabled(self):
        user = UserFactory()
        ma = UserAdmin(User, AdminSite())
        self.assertFalse('native_name' in ma.get_list_display(request))
        self.assertFalse('native_name' in ma.get_search_fields(request))
        self.assertTrue(all('native_name' not in fieldset[1]['fields']
                            for fieldset in ma.get_fieldsets(request, user)))

    @override_waldur_core_settings(NATIVE_NAME_ENABLED=True)
    def test_native_name_is_rendered_in_user_admin_if_feature_is_enabled(self):
        user = UserFactory()
        ma = UserAdmin(User, AdminSite())
        self.assertTrue('native_name' in ma.get_list_display(request))
        self.assertTrue('native_name' in ma.get_search_fields(request))
        self.assertTrue(any('native_name' in fieldset[1]['fields']
                            for fieldset in ma.get_fieldsets(request, user)))

    @override_waldur_core_settings(NATIVE_NAME_ENABLED=False)
    def test_native_name_is_omitted_in_customer_admin_if_feature_is_disabled(self):
        customer = CustomerFactory()
        ma = CustomerAdmin(Customer, AdminSite())
        self.assertFalse('native_name' in ma.get_fields(request, customer))

    @override_waldur_core_settings(NATIVE_NAME_ENABLED=True)
    def test_native_name_is_rendered_in_customer_admin_if_feature_is_enabled(self):
        customer = CustomerFactory()
        ma = CustomerAdmin(Customer, AdminSite())
        self.assertTrue('native_name' in ma.get_fields(request, customer))



from django.test import TestCase
from django.utils import timezone
from freezegun import freeze_time

from waldur_core.structure.tests import factories
from waldur_core.structure.tests import models as test_models


class ServiceSettingsSignalsTest(TestCase):

    def setUp(self):
        self.shared_service_settings = factories.ServiceSettingsFactory(shared=True)

    def test_shared_service_is_created_for_new_customer(self):
        customer = factories.CustomerFactory()

        self.assertTrue(test_models.TestService.objects.filter(
            customer=customer, settings=self.shared_service_settings, available_for_all=True).exists())


class ServiceProjectLinkSignalsTest(TestCase):

    def test_new_project_connects_to_available_services_of_customer(self):
        customer = factories.CustomerFactory()
        service = self.create_service(customer, available_for_all=True)

        other_customer = factories.CustomerFactory()
        other_service = self.create_service(other_customer, available_for_all=True)

        # Act
        project = factories.ProjectFactory(customer=customer)

        # Assert
        self.assertTrue(self.link_exists(project, service))
        self.assertFalse(self.link_exists(project, other_service))

    def test_if_service_became_available_it_connects_to_all_projects_of_customer(self):
        customer = factories.CustomerFactory()
        service = self.create_service(customer, available_for_all=False)
        project = factories.ProjectFactory(customer=customer)

        other_customer = factories.CustomerFactory()
        other_project = factories.ProjectFactory(customer=other_customer)

        # Act
        service.available_for_all = True
        service.save()

        # Assert
        self.assertTrue(self.link_exists(project, service))
        self.assertFalse(self.link_exists(other_project, service))

    def create_service(self, customer, available_for_all):
        service_settings = factories.ServiceSettingsFactory(shared=False)
        return test_models.TestService.objects.create(customer=customer,
                                                      settings=service_settings,
                                                      available_for_all=available_for_all)

    def link_exists(self, project, service):
        return test_models.TestServiceProjectLink.objects.filter(
            project=project, service=service).exists()


class StartTimeTest(TestCase):
    def test_if_resource_becomes_online_start_time_is_initialized(self):
        now = timezone.now()
        with freeze_time(now):
            vm = factories.TestNewInstanceFactory(runtime_state='in-progress', start_time=None)
            vm.runtime_state = 'online'
            vm.save()
            vm.refresh_from_db()
            self.assertEqual(vm.start_time, now)

    def test_if_resource_becomes_offline_start_time_is_resetted(self):
        vm = factories.TestNewInstanceFactory(runtime_state='paused', start_time=timezone.now())
        vm.runtime_state = 'offline'
        vm.save()
        vm.refresh_from_db()
        self.assertEqual(vm.start_time, None)

    def test_if_runtime_state_changed_to_other_state_start_time_is_not_modified(self):
        vm = factories.TestNewInstanceFactory(runtime_state='online', start_time=None)
        vm.runtime_state = 'extending'
        vm.save()
        vm.refresh_from_db()
        self.assertEqual(vm.start_time, None)



from . import views


def register_in(router):
    router.register(r'resource-sla-state-transition', views.ResourceSlaStateTransitionViewSet,
                    base_name='resource-sla-state-transition')



from django.db import models as django_models

from waldur_core.core import models as core_models
from waldur_core.quotas import fields, models as quotas_models


class GrandparentModel(core_models.UuidMixin, quotas_models.QuotaModelMixin, core_models.DescendantMixin):
    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        regular_quota = fields.QuotaField()
        quota_with_default_limit = fields.QuotaField(default_limit=100)
        usage_aggregator_quota = fields.UsageAggregatorQuotaField(
            get_children=lambda scope: ChildModel.objects.filter(parent__parent=scope),
        )
        limit_aggregator_quota = fields.LimitAggregatorQuotaField(
            get_children=lambda scope: ChildModel.objects.filter(parent__parent=scope),
        )

    regular_quota = fields.QuotaLimitField(quota_field=Quotas.regular_quota)


class ParentModel(core_models.UuidMixin, quotas_models.QuotaModelMixin, core_models.DescendantMixin):
    parent = django_models.ForeignKey(GrandparentModel, related_name='children')

    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        counter_quota = fields.CounterQuotaField(
            target_models=lambda: [ChildModel],
            path_to_scope='parent',
        )
        two_targets_counter_quota = fields.CounterQuotaField(
            target_models=lambda: [ChildModel, SecondChildModel],
            path_to_scope='parent',
        )
        delta_quota = fields.CounterQuotaField(
            target_models=lambda: [ChildModel],
            path_to_scope='parent',
            get_delta=lambda scope: 10
        )
        usage_aggregator_quota = fields.UsageAggregatorQuotaField(
            get_children=lambda scope: scope.children.all(),
        )
        limit_aggregator_quota = fields.LimitAggregatorQuotaField(
            get_children=lambda scope: scope.children.all(),
            default_limit=0,
        )
        second_usage_aggregator_quota = fields.UsageAggregatorQuotaField(
            get_children=lambda scope: scope.children.all(),
            child_quota_name='usage_aggregator_quota',
        )
        total_quota = fields.TotalQuotaField(
            target_models=lambda: [SecondChildModel],
            path_to_scope='parent',
            target_field='size',
        )

    def get_parents(self):
        return [self.parent]


class NonQuotaParentModel(core_models.UuidMixin, core_models.DescendantMixin):
    """ Allow to make sure that quotas propagation do not fail if quota scope has parent without quotas """
    pass


class ChildModel(core_models.UuidMixin, quotas_models.QuotaModelMixin, core_models.DescendantMixin):
    parent = django_models.ForeignKey(ParentModel, related_name='children')
    non_quota_parent = django_models.ForeignKey(NonQuotaParentModel, related_name='children', blank=True, null=True)

    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        regular_quota = fields.QuotaField()
        usage_aggregator_quota = fields.QuotaField()  # this quota is aggregated by parent and grandparent
        limit_aggregator_quota = fields.QuotaField(default_limit=0)  # this quota is aggregated by parent and grandparent

    def get_parents(self):
        if self.non_quota_parent:
            return [self.parent, self.non_quota_parent]
        return [self.parent]


class SecondChildModel(core_models.UuidMixin, quotas_models.QuotaModelMixin, core_models.DescendantMixin):
    parent = django_models.ForeignKey(ParentModel, related_name='second_children')
    size = django_models.IntegerField(default=0)

    def get_parents(self):
        return [self.parent]




from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import resolve
from rest_framework.test import APIRequestFactory
from six.moves.urllib.parse import urlparse

from .. import factories as structure_factories
from ...serializers import BasicUserSerializer

User = get_user_model()


class UUIDSerializerTest(TestCase):
    def setUp(self):
        factory = APIRequestFactory()
        request = factory.get('/users/')
        context = {'request': request}
        user = structure_factories.UserFactory()
        serializer = BasicUserSerializer(instance=user, context=context)
        self.data = serializer.data

    def test_url_and_uuid_do_not_contain_hyphenation(self):
        path = urlparse(self.data['url']).path
        match = resolve(path)
        self.assertEqual(match.url_name, 'user-detail')

        value = match.kwargs.get('uuid')
        self.assertEqual(value, self.data['uuid'])
        self.assertTrue('-' not in value)



from rest_framework import serializers

from waldur_core.structure import serializers as structure_serializers
from . import models


class ServiceSerializer(structure_serializers.BaseServiceSerializer):
    SERVICE_ACCOUNT_EXTRA_FIELDS = {
        'tenant_name': '',
        'availability_zone': '',
    }

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.TestService
        required_fields = 'backend_url', 'username', 'password'


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):
    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.TestServiceProjectLink
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'test-detail'},
        }


class NewInstanceSerializer(structure_serializers.VirtualMachineSerializer):

    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='test-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='test-spl-detail',
        queryset=models.TestServiceProjectLink.objects.all())

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.TestNewInstance



from django.contrib.admin.models import LogEntry
from django.core.management.base import BaseCommand

from waldur_core.cost_tracking import models as cost_tracking_models


class Command(BaseCommand):
    help = "Remove instances that have FK to stale content types."

    def handle(self, *args, **options):
        for estimate in cost_tracking_models.PriceEstimate.objects.all():
            if estimate.content_type.model_class() is None:
                estimate.delete()

        for item in cost_tracking_models.DefaultPriceListItem.objects.all():
            if item.resource_content_type.model_class() is None:
                item.delete()

        for entry in LogEntry.objects.all():
            if entry.content_type.model_class() is None:
                entry.delete()



from __future__ import unicode_literals

import unittest

from waldur_core.core import utils


class TestFormatTimeAndValueToSegmentList(unittest.TestCase):

    def test_function_splits_time_to_segments_right(self):
        start_timestamp = 100
        end_timestamp = 200
        segments_count = 5
        # when
        segment_list = utils.format_time_and_value_to_segment_list(
            [], segments_count, start_timestamp, end_timestamp)
        # then
        self.assertEqual(len(segment_list), segments_count)
        step = (end_timestamp - start_timestamp) / segments_count
        for index, segment in enumerate(segment_list):
            self.assertEqual(segment['from'], start_timestamp + step * index)
            self.assertEqual(segment['to'], start_timestamp + step * (index + 1))

    def test_function_sums_values_in_segments_right(self):
        start_timestamp = 20
        end_timestamp = 60
        segments_count = 2
        first_segment_time_value_list = [(22, 1), (23, 2)]
        second_segment_time_value_list = [(52, 2), (59, 2), (43, 3)]
        time_and_value_list = first_segment_time_value_list + second_segment_time_value_list
        # when
        segment_list = utils.format_time_and_value_to_segment_list(
            time_and_value_list, segments_count, start_timestamp, end_timestamp)
        # then
        first_segment, second_segment = segment_list
        expected_first_segment_value = sum([value for _, value in first_segment_time_value_list])
        expected_second_segment_value = sum([value for _, value in second_segment_time_value_list])
        self.assertEqual(first_segment['value'], expected_first_segment_value)
        self.assertEqual(second_segment['value'], expected_second_segment_value)



import mock
from celery.app.task import Context
from celery.backends.base import Backend
from django.test import testcases


class ExecutorTest(testcases.TestCase):
    def setUp(self):
        app = mock.Mock(**{
            'conf.result_serializer': 'json',
            'conf.accept_content': None
        })
        self.backend = Backend(app)
        errback = {"chord_size": None,
                   "task": "waldur_core.core.tasks.ErrorStateTransitionTask",
                   "args": ["waldur.obj:1"],
                   "immutable": False,
                   "subtask_type": None,
                   "kwargs": {},
                   "options": {}
                   }
        self.request = Context(errbacks=[errback], id='task_id', root_id='root_id')

    @mock.patch('waldur_core.core.tasks.group')
    def test_use_old_signature_in_task_error(self, mock_group):
        self.backend._call_task_errbacks(self.request, Exception('test'), '')
        self.assertEqual(mock_group.call_count, 1)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields
import waldur_core.structure.images
import django.db.models.deletion
import waldur_core.core.fields
import django.core.validators
import django_fsm
import waldur_core.core.validators


class Migration(migrations.Migration):

    #replaces = [('structure', '0001_initial'), ('structure', '0002_customer_native_name'), ('structure', '0003_protect_non_empty_customers'), ('structure', '0004_init_new_quotas'), ('structure', '0005_init_customers_quotas'), ('structure', '0006_inherit_namemixin'), ('structure', '0007_add_service_model'), ('structure', '0008_add_customer_billing_fields'), ('structure', '0009_update_service_models'), ('structure', '0010_add_oracle_service_type'), ('structure', '0011_customer_registration_code'), ('structure', '0012_customer_image'), ('structure', '0013_servicesettings_customer'), ('structure', '0014_servicesettings_options'), ('structure', '0015_drop_service_polymorphic'), ('structure', '0016_init_nc_resource_count_quotas'), ('structure', '0017_add_azure_service_type'), ('structure', '0018_service_settings_plural_form'), ('structure', '0019_rename_nc_service_count_to_nc_service_project_link_count'), ('structure', '0020_servicesettings_certificate'), ('structure', '0021_balancehistory')]

    dependencies = [
        ('auth', '0001_initial'),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Customer',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('native_name', models.CharField(default='', max_length=160, blank=True)),
                ('abbreviation', models.CharField(max_length=8, blank=True)),
                ('contact_details', models.TextField(blank=True, validators=[django.core.validators.MaxLengthValidator(500)])),
                ('billing_backend_id', models.CharField(max_length=255, blank=True)),
                ('balance', models.DecimalField(null=True, max_digits=9, decimal_places=3, blank=True)),
                ('image', models.ImageField(null=True, upload_to=waldur_core.structure.images.get_upload_path, blank=True)),
                ('registration_code', models.CharField(default='', max_length=160, blank=True)),
            ],
            options={
                'abstract': False,
                'verbose_name': 'organization',
            },
        ),
        migrations.CreateModel(
            name='CustomerRole',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('role_type', models.SmallIntegerField(choices=[(0, 'Owner')])),
                ('customer', models.ForeignKey(verbose_name='organization', related_name='roles', to='structure.Customer')),
                ('permission_group', models.OneToOneField(to='auth.Group')),
            ],
            options={
                'unique_together': set([('customer', 'role_type')]),
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('customer', models.ForeignKey(verbose_name='organization', related_name='projects', on_delete=django.db.models.deletion.PROTECT, to='structure.Customer')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProjectRole',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('role_type', models.SmallIntegerField(choices=[(0, 'Administrator'), (1, 'Manager')])),
                ('permission_group', models.OneToOneField(to='auth.Group')),
                ('project', models.ForeignKey(related_name='roles', to='structure.Project')),
            ],
            options={
                'unique_together': set([('project', 'role_type')]),
            },
        ),
        migrations.CreateModel(
            name='ProjectGroup',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('customer', models.ForeignKey(verbose_name='organization', related_name='project_groups', on_delete=django.db.models.deletion.PROTECT, to='structure.Customer')),
                ('projects', models.ManyToManyField(related_name='project_groups', to='structure.Project')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProjectGroupRole',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('role_type', models.SmallIntegerField(choices=[(0, 'Group Manager')])),
                ('permission_group', models.OneToOneField(to='auth.Group')),
                ('project_group', models.ForeignKey(related_name='roles', to='structure.ProjectGroup')),
            ],
            options={
                'unique_together': set([('project_group', 'role_type')]),
            },
        ),
        migrations.CreateModel(
            name='ServiceSettings',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('customer', models.ForeignKey(verbose_name='organization', related_name='service_settings', blank=True, to='structure.Customer', null=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('state', django_fsm.FSMIntegerField(default=1, choices=[(1, 'Sync Scheduled'), (2, 'Syncing'), (3, 'In Sync'), (4, 'Erred')])),
                ('backend_url', waldur_core.core.fields.BackendURLField(null=True, blank=True)),
                ('username', models.CharField(max_length=100, null=True, blank=True)),
                ('password', models.CharField(max_length=100, null=True, blank=True)),
                ('certificate', models.FileField(blank=True, null=True, upload_to='certs', validators=[waldur_core.core.validators.FileTypeValidator(allowed_extensions=['pem'], allowed_types=['application/x-pem-file', 'application/x-x509-ca-cert', 'text/plain'])])),
                ('token', models.CharField(max_length=255, null=True, blank=True)),
                ('type', models.SmallIntegerField(choices=[(1, b'OpenStack'), (2, b'DigitalOcean'), (3, b'Amazon'), (4, b'Jira'), (5, b'GitLab'), (6, b'Oracle'), (7, b'Azure')])),
                ('options', waldur_core.core.fields.JSONField(help_text='Extra options', blank=True)),
                ('shared', models.BooleanField(default=False, help_text='Anybody can use it')),
                ('dummy', models.BooleanField(default=False, help_text='Emulate backend operations')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Service settings',
                'verbose_name_plural': 'Service settings',
            },
        ),
        migrations.CreateModel(
            name='BalanceHistory',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False)),
                ('amount', models.DecimalField(max_digits=9, decimal_places=3)),
                ('customer', models.ForeignKey(verbose_name='organization', to='structure.Customer')),
            ],
        ),
    ]



from django.test.utils import override_settings
from django.urls import reverse
from rest_framework import status
from rest_framework import test

from waldur_core.core.tests.helpers import override_waldur_core_settings
from waldur_core.structure.images import dummy_image
from waldur_core.structure.models import CustomerRole
from waldur_core.structure.tests.factories import UserFactory, CustomerFactory


@override_settings(MEDIA_URL='/media/')
class ImageUploadTest(test.APITransactionTestCase):
    def setUp(self):
        self.staff = UserFactory(is_staff=True)
        self.owner = UserFactory()
        self.user = UserFactory()
        self.customer = CustomerFactory()
        self.customer.add_user(self.owner, CustomerRole.OWNER)
        self.url = reverse('customer_image', kwargs={'uuid': self.customer.uuid.hex})

    def test_default_customer_logo(self):
        self.client.force_authenticate(user=self.staff)
        self.assert_default_logo()

        with dummy_image() as image:
            self.assert_can_upload_image(image)
            self.assert_can_delete_image()

        self.assert_default_logo()

    # Positive
    def test_staff_can_upload_and_delete_customer_logo(self):
        self.client.force_authenticate(user=self.staff)

        with dummy_image() as image:
            self.assert_can_upload_image(image)
            self.assert_can_delete_image()

    def test_customer_owner_can_upload_and_delete_customer_logo(self):
        self.client.force_authenticate(user=self.owner)

        with dummy_image() as image:
            self.assert_can_upload_image(image)
            self.assert_can_delete_image()

    # Negative
    def test_user_cannot_upload_logo_for_customer_he_is_not_owner_of(self):
        self.client.force_authenticate(user=self.user)

        with dummy_image() as image:
            response = self.upload_image(self.url, image)
            self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)

    # Helpers
    def assert_can_upload_image(self, image):
        response = self.upload_image(self.url, image)
        self.assertEqual(status.HTTP_200_OK, response.status_code, response.data)
        self.assertIn('image', response.data)
        self.assertIsNotNone(response.data['image'])

    def assert_can_delete_image(self):
        response = self.client.delete(self.url)
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code, response.data)

    @override_waldur_core_settings(DEFAULT_CUSTOMER_LOGO='default_logo.jpg')
    def assert_default_logo(self):
        url = reverse('customer-detail', kwargs={'uuid': self.customer.uuid.hex})
        response = self.client.get(url)

        self.assertEqual(status.HTTP_200_OK, response.status_code)
        self.assertEqual('default_logo.jpg', response.data['image'])

    def upload_image(self, url, image):
        return self.client.put(self.url, {'image': image}, format='multipart')



from django.db import transaction
from django.db.models import signals

from waldur_core.quotas import models, utils, fields
from waldur_core.quotas.exceptions import CreationConditionFailedQuotaError


# XXX: rewrite global quotas
def create_global_quotas(**kwargs):
    for model in utils.get_models_with_quotas():
        if hasattr(model, 'GLOBAL_COUNT_QUOTA_NAME'):
            models.Quota.objects.get_or_create(name=getattr(model, 'GLOBAL_COUNT_QUOTA_NAME'))


def increase_global_quota(sender, instance=None, created=False, **kwargs):
    if created and hasattr(sender, 'GLOBAL_COUNT_QUOTA_NAME'):
        with transaction.atomic():
            global_quota = models.Quota.objects.select_for_update().get(
                name=getattr(sender, 'GLOBAL_COUNT_QUOTA_NAME'))
            global_quota.usage += 1
            global_quota.save()


def decrease_global_quota(sender, **kwargs):
    if hasattr(sender, 'GLOBAL_COUNT_QUOTA_NAME'):
        with transaction.atomic():
            global_quota = models.Quota.objects.select_for_update().get(
                name=getattr(sender, 'GLOBAL_COUNT_QUOTA_NAME'))
            global_quota.usage -= 1
            global_quota.save()


# new quotas

def init_quotas(sender, instance, created=False, **kwargs):
    """ Initialize new instances quotas """
    if not created:
        return
    for field in sender.get_quotas_fields():
        try:
            field.get_or_create_quota(scope=instance)
        except CreationConditionFailedQuotaError:
            pass


def count_quota_handler_factory(count_quota_field):
    """ Creates handler that will recalculate count_quota on creation/deletion """

    def recalculate_count_quota(sender, instance, **kwargs):
        signal = kwargs['signal']
        if signal == signals.post_save and kwargs.get('created'):
            count_quota_field.add_usage(instance, delta=1)
        elif signal == signals.post_delete:
            count_quota_field.add_usage(instance, delta=-1, fail_silently=True)

    return recalculate_count_quota


def handle_aggregated_quotas(sender, instance, **kwargs):
    """ Call aggregated quotas fields update methods """
    quota = instance
    # aggregation is not supported for global quotas.
    if quota.scope is None:
        return
    quota_field = quota.get_field()
    # usage aggregation should not count another usage aggregator field to avoid calls duplication.
    if isinstance(quota_field, fields.UsageAggregatorQuotaField) or quota_field is None:
        return
    signal = kwargs['signal']
    for aggregator_quota in quota_field.get_aggregator_quotas(quota):
        field = aggregator_quota.get_field()
        if signal == signals.post_save:
            field.post_child_quota_save(aggregator_quota.scope, child_quota=quota, created=kwargs.get('created'))
        elif signal == signals.pre_delete:
            field.pre_child_quota_delete(aggregator_quota.scope, child_quota=quota)



from ddt import ddt, data
from django.db import models
from rest_framework import status, test

from waldur_core.core.tests.helpers import override_waldur_core_settings
from waldur_core.structure.models import ProjectRole
from waldur_core.structure.tests import factories, fixtures, models as test_models


@ddt
class ServiceListTest(test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ServiceFixture()
        self.service = self.fixture.service
        self.spl = self.fixture.service_project_link

    def get_list(self):
        return self.client.get(factories.TestServiceFactory.get_list_url())

    def test_anonymous_user_cannot_list_services(self):
        response = self.get_list()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    @data('staff', 'owner', 'admin', 'manager')
    def test_authorized_user_can_list_services(self, user):
        self.client.force_authenticate(user=getattr(self.fixture, user))

        response = self.get_list()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        service_url = factories.TestServiceFactory.get_url(self.service)
        self.assertIn(service_url, [instance['url'] for instance in response.data])

    def test_user_cannot_list_services_of_projects_he_has_no_role_in(self):
        self.client.force_authenticate(user=self.fixture.user)

        response = self.get_list()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(0, len(response.data))


@ddt
class ServiceCreateTest(test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ProjectFixture()
        self.url = factories.TestServiceFactory.get_list_url()

    def test_if_required_fields_are_specified_service_is_created(self):
        self.client.force_authenticate(self.fixture.owner)
        response = self.client.post(self.url, self._get_valid_payload())
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_if_required_field_is_not_specified_error_raised(self):
        self.client.force_authenticate(self.fixture.owner)

        payload = self._get_valid_payload()
        del payload['username']
        response = self.client.post(self.url, payload)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn('username', response.data)

    @data('staff', 'owner')
    def test_staff_and_owner_can_create_service(self, user):
        self.assert_user_can_create_service(user)

    @data('admin', 'manager')
    def test_admin_and_manager_can_not_create_service(self, user):
        self.assert_user_can_not_create_service(user)

    @override_waldur_core_settings(ONLY_STAFF_MANAGES_SERVICES=True)
    def test_if_only_staff_manages_services_he_can_create_it(self):
        self.assert_user_can_create_service('staff')

    @data('owner', 'admin', 'manager')
    @override_waldur_core_settings(ONLY_STAFF_MANAGES_SERVICES=True)
    def test_if_only_staff_manages_services_other_users_can_not_create_it(self, user):
        self.assert_user_can_not_create_service(user)

    def assert_user_can_create_service(self, user):
        self.client.force_authenticate(getattr(self.fixture, user))
        response = self.client.post(self.url, self._get_valid_payload())
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def assert_user_can_not_create_service(self, user):
        self.client.force_authenticate(getattr(self.fixture, user))
        response = self.client.post(self.url, self._get_valid_payload())
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def _get_valid_payload(self):
        customer_url = factories.CustomerFactory.get_url(self.fixture.customer)
        return {
            'name': 'Test service',
            'customer': customer_url,
            'backend_url': 'http://example.com/',
            'username': 'admin',
            'password': 'secret',
        }


@ddt
class ServiceDeleteTest(test.APITransactionTestCase):
    def setUp(self):
        self.fixture = fixtures.ServiceFixture()
        self.service = self.fixture.service
        self.url = factories.TestServiceFactory.get_url(self.service)

    @data('staff', 'owner')
    def test_staff_and_owner_can_delete_service(self, user):
        self.assert_user_can_delete_service(user)

    @data('admin', 'manager')
    def test_admin_and_manager_can_not_delete_service(self, user):
        self.assert_user_can_not_delete_service(user, status.HTTP_404_NOT_FOUND)

    @data('staff')
    @override_waldur_core_settings(ONLY_STAFF_MANAGES_SERVICES=True)
    def test_if_only_staff_manages_services_he_can_delete_it(self, user):
        self.assert_user_can_delete_service(user)

    @override_waldur_core_settings(ONLY_STAFF_MANAGES_SERVICES=True)
    def test_if_only_staff_manages_services_owner_can_not_delete_it(self):
        self.assert_user_can_not_delete_service('owner', status.HTTP_403_FORBIDDEN)

    def assert_user_can_delete_service(self, user):
        self.client.force_authenticate(getattr(self.fixture, user))
        response = self.client.delete(self.url)
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(test_models.TestService.objects.filter(pk=self.service.pk).exists())

    def assert_user_can_not_delete_service(self, user, expected_status):
        self.client.force_authenticate(getattr(self.fixture, user))
        response = self.client.delete(self.url)
        self.assertEqual(response.status_code, expected_status)
        self.assertTrue(test_models.TestService.objects.filter(pk=self.service.pk).exists())


class ServiceResourcesCounterTest(test.APITransactionTestCase):
    """
    There's one shared service. Also there are 2 users each of which has one project.
    There's one VM in each project. Service counters for each user should equal 1.
    For staff user resource counter should equal 2.
    """

    def setUp(self):
        self.customer = factories.CustomerFactory()
        self.settings = factories.ServiceSettingsFactory(shared=True)
        self.service = factories.TestServiceFactory(customer=self.customer, settings=self.settings)

        self.user1 = factories.UserFactory()
        self.project1 = factories.ProjectFactory(customer=self.customer)
        self.project1.add_user(self.user1, ProjectRole.ADMINISTRATOR)
        self.spl1 = factories.TestServiceProjectLinkFactory(service=self.service, project=self.project1)
        self.vm1 = factories.TestNewInstanceFactory(service_project_link=self.spl1)

        self.user2 = factories.UserFactory()
        self.project2 = factories.ProjectFactory(customer=self.customer)
        self.project2.add_user(self.user2, ProjectRole.ADMINISTRATOR)
        self.spl2 = factories.TestServiceProjectLinkFactory(service=self.service, project=self.project2)
        self.vm2 = factories.TestNewInstanceFactory(service_project_link=self.spl2)

        self.service_url = factories.TestServiceFactory.get_url(self.service)

    def test_counters_for_shared_providers_should_be_filtered_by_user(self):
        self.client.force_authenticate(self.user1)
        response = self.client.get(self.service_url)
        self.assertEqual(1, response.data['resources_count'])

        self.client.force_authenticate(self.user2)
        response = self.client.get(self.service_url)
        self.assertEqual(1, response.data['resources_count'])

    def test_counters_are_not_filtered_for_staff(self):
        self.client.force_authenticate(factories.UserFactory(is_staff=True))
        response = self.client.get(self.service_url)
        self.assertEqual(2, response.data['resources_count'])

    def test_subresources_are_skipped(self):
        factories.TestSubResourceFactory(service_project_link=self.spl1)
        self.client.force_authenticate(self.user1)
        response = self.client.get(self.service_url)
        self.assertEqual(1, response.data['resources_count'])


class ServiceUnlinkTest(test.APITransactionTestCase):
    def test_when_service_is_unlinked_all_related_resources_are_unlinked_too(self):
        resource = factories.TestNewInstanceFactory()
        service = resource.service_project_link.service
        unlink_url = factories.TestServiceFactory.get_url(service, 'unlink')

        self.client.force_authenticate(factories.UserFactory(is_staff=True))
        response = self.client.post(unlink_url)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertRaises(models.ObjectDoesNotExist, service.refresh_from_db)

    def test_owner_can_unlink_managed_service(self):
        fixture = fixtures.ServiceFixture()
        service = fixture.service

        unlink_url = factories.TestServiceFactory.get_url(service, 'unlink')
        self.client.force_authenticate(fixture.owner)
        response = self.client.post(unlink_url)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertRaises(models.ObjectDoesNotExist, service.refresh_from_db)

    def test_owner_cannot_unlink_service_with_shared_settings(self):
        fixture = fixtures.ServiceFixture()
        service_settings = factories.ServiceSettingsFactory(shared=True)
        service = test_models.TestService.objects.get(customer=fixture.customer, settings=service_settings)
        unlink_url = factories.TestServiceFactory.get_url(service, 'unlink')
        self.client.force_authenticate(fixture.owner)

        response = self.client.post(unlink_url)

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertTrue(test_models.TestService.objects.filter(pk=service.pk).exists())

    @override_waldur_core_settings(ONLY_STAFF_MANAGES_SERVICES=True)
    def test_if_only_staff_manages_services_owner_cannot_unlink_it(self):
        fixture = fixtures.ServiceFixture()
        service = fixture.service

        unlink_url = factories.TestServiceFactory.get_url(service, 'unlink')
        self.client.force_authenticate(fixture.owner)
        response = self.client.post(unlink_url)

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertTrue(test_models.TestService.objects.filter(pk=service.pk).exists())



from collections import OrderedDict

from django.utils.encoding import force_text
from django.utils.http import urlencode
from rest_framework import exceptions
from rest_framework.metadata import SimpleMetadata
from rest_framework.request import clone_request
from rest_framework.reverse import reverse

from waldur_core.core.utils import sort_dict


class ActionSerializer(object):
    def __init__(self, func, name, request, view, resource):
        self.func = func
        self.name = name
        self.request = request
        self.resource = resource
        self.view = view

    def serialize(self):
        reason = self.get_reason()

        return {
            'title': self.get_title(),
            'method': self.get_method(),
            'destructive': self.is_destructive(),
            'url': self.get_url(),
            'reason': reason,
            'enabled': not reason
        }

    def is_destructive(self):
        if self.name == 'destroy':
            return True
        return getattr(self.func, 'destructive', False)

    def get_title(self):
        try:
            return getattr(self.func, 'title')
        except AttributeError:
            return self.name.replace('_', ' ').title()

    def get_reason(self):
        try:
            self.view.initial(self.request)
        except exceptions.APIException as e:
            return force_text(e)

    def get_method(self):
        if self.name == 'destroy':
            return 'DELETE'
        elif self.name == 'update':
            return 'PUT'
        return getattr(self.func, 'method', 'POST')

    def get_url(self):
        base_url = self.request.build_absolute_uri()
        method = self.get_method()
        if method in ('DELETE', 'PUT'):
            return base_url
        return base_url + self.name + '/'


class ActionsMetadata(SimpleMetadata):
    """
    Difference from SimpleMetadata class:
    1) Skip read-only fields, because options are used only for provisioning new resource.
    2) Don't expose choices for fields with queryset in order to reduce size of response.
    3) Attach actions metadata
    """

    def determine_metadata(self, request, view):
        self.request = request
        metadata = OrderedDict()
        if view.lookup_field in view.kwargs:
            metadata['actions'] = self.get_actions(request, view)
        else:
            metadata['actions'] = self.determine_actions(request, view)
        return metadata

    def get_actions(self, request, view):
        """
        Return metadata for resource-specific actions,
        such as start, stop, unlink
        """
        metadata = OrderedDict()
        actions = self.get_resource_actions(view)

        resource = view.get_object()
        for action_name, action in actions.items():
            if action_name == 'update':
                view.request = clone_request(request, 'PUT')
            else:
                view.action = action_name

            data = ActionSerializer(action, action_name, request, view, resource)
            metadata[action_name] = data.serialize()
            if not metadata[action_name]['enabled']:
                continue
            fields = self.get_action_fields(view, action_name, resource)
            if not fields:
                metadata[action_name]['type'] = 'button'
            else:
                metadata[action_name]['type'] = 'form'
                metadata[action_name]['fields'] = fields

            view.action = None
            view.request = request

        return metadata

    @classmethod
    def get_resource_actions(cls, view):
        actions = {}
        disabled_actions = getattr(view.__class__, 'disabled_actions', [])

        for key in dir(view.__class__):
            callback = getattr(view.__class__, key)
            if getattr(callback, 'deprecated', False):
                continue
            if 'post' not in getattr(callback, 'bind_to_methods', []):
                continue
            if key in disabled_actions:
                continue
            actions[key] = callback

        if 'DELETE' in view.allowed_methods and 'destroy' not in disabled_actions:
            actions['destroy'] = view.destroy

        if 'PUT' in view.allowed_methods and 'update' not in disabled_actions:
            actions['update'] = view.update

        return sort_dict(actions)

    def get_action_fields(self, view, action_name, resource):
        """
        Get fields exposed by action's serializer
        """
        serializer = view.get_serializer(resource)
        fields = OrderedDict()
        if not isinstance(serializer, view.serializer_class) or action_name == 'update':
            fields = self.get_fields(serializer.fields)
        return fields

    def get_serializer_info(self, serializer):
        """
        Given an instance of a serializer, return a dictionary of metadata
        about its fields.
        """
        if hasattr(serializer, 'child'):
            # If this is a `ListSerializer` then we want to examine the
            # underlying child serializer instance instead.
            serializer = serializer.child
        return self.get_fields(serializer.fields)

    def get_fields(self, serializer_fields):
        """
        Get fields metadata skipping empty fields
        """
        fields = OrderedDict()
        for field_name, field in serializer_fields.items():
            # Skip tags field in action because it is needed only for resource creation
            # See also: WAL-1223
            if field_name == 'tags':
                continue
            info = self.get_field_info(field, field_name)
            if info:
                fields[field_name] = info
        return fields

    def get_field_info(self, field, field_name):
        """
        Given an instance of a serializer field, return a dictionary
        of metadata about it.
        """
        field_info = OrderedDict()
        field_info['type'] = self.label_lookup[field]
        field_info['required'] = getattr(field, 'required', False)

        attrs = [
            'label', 'help_text', 'default_value', 'placeholder', 'required',
            'min_length', 'max_length', 'min_value', 'max_value', 'many'
        ]

        if getattr(field, 'read_only', False):
            return None

        for attr in attrs:
            value = getattr(field, attr, None)
            if value is not None and value != '':
                field_info[attr] = force_text(value, strings_only=True)

        if 'label' not in field_info:
            field_info['label'] = field_name.replace('_', ' ').title()

        if hasattr(field, 'view_name'):
            list_view = field.view_name.replace('-detail', '-list')
            base_url = reverse(list_view, request=self.request)
            field_info['type'] = 'select'
            field_info['url'] = base_url
            if hasattr(field, 'query_params'):
                field_info['url'] += '?%s' % urlencode(field.query_params)
            field_info['value_field'] = getattr(field, 'value_field', 'url')
            field_info['display_name_field'] = getattr(field, 'display_name_field', 'display_name')

        if hasattr(field, 'choices') and not hasattr(field, 'queryset'):
            field_info['choices'] = [
                {
                    'value': choice_value,
                    'display_name': force_text(choice_name, strings_only=True)
                }
                for choice_value, choice_name in field.choices.items()
            ]

        return field_info



import logging
import time

from django.conf import settings
from django.core import mail
from rest_framework import test
from six.moves import mock

from waldur_core.logging import models as logging_models
from waldur_core.logging.log import HookHandler
from waldur_core.logging.tasks import process_event
from waldur_core.structure import models as structure_models
from waldur_core.structure.log import event_logger
from waldur_core.structure.tests import factories as structure_factories


class TestHookService(test.APITransactionTestCase):
    def setUp(self):
        self.owner = structure_factories.UserFactory()
        self.customer = structure_factories.CustomerFactory()
        self.customer.add_user(self.owner, structure_models.CustomerRole.OWNER)
        self.other_user = structure_factories.UserFactory()

        self.event_type = 'customer_update_succeeded'
        self.other_event = 'customer_deletion_succeeded'
        self.message = 'Customer {customer_name} has been updated.'
        self.event = {
            'message': self.message,
            'type': self.event_type,
            'context': event_logger.customer.compile_context(customer=self.customer),
            'timestamp': time.time()
        }

        # Create email hook for another user
        self.other_hook = logging_models.EmailHook.objects.create(user=self.other_user,
                                                                  email=self.owner.email,
                                                                  event_types=[self.event_type])

    @mock.patch('celery.app.base.Celery.send_task')
    def test_logger_handler_sends_task_if_handler_attached(self, mocked_task):
        # Prepare logger
        logger = logging.getLogger('waldur_core')
        logger.setLevel(logging.DEBUG)

        # Inject handler
        handler = HookHandler()
        logger.addHandler(handler)

        event_logger.customer.warning(self.message,
                                      event_type=self.event_type,
                                      event_context={'customer': self.customer})

        mocked_task.assert_called_once_with('waldur_core.logging.process_event', mock.ANY, {}, countdown=2)
        mocked_task.reset_mock()

        # Remove hook handler so that other tests won't depend on it
        logger.removeHandler(handler)

        # Trigger an event
        event_logger.customer.warning(self.message,
                                      event_type=self.event_type,
                                      event_context={'customer': self.customer})

        # If hook handler is not attached hook is not processed
        self.assertFalse(mocked_task.called)

    def test_email_hook_filters_events_by_user_and_event_type(self):
        # Create email hook for customer owner
        email_hook = logging_models.EmailHook.objects.create(user=self.owner,
                                                             email=self.owner.email,
                                                             event_types=[self.event_type])

        # Trigger processing
        process_event(self.event)

        # Test that one message has been sent for email hook of customer owner
        self.assertEqual(len(mail.outbox), 1)

        # Verify that destination address of message is correct
        self.assertEqual(mail.outbox[0].to, [email_hook.email])

    @mock.patch('requests.post')
    def test_webhook_makes_post_request_against_destination_url(self, requests_post):

        # Create web hook for customer owner
        self.web_hook = logging_models.WebHook.objects.create(user=self.owner,
                                                              destination_url='http://example.com/',
                                                              event_types=[self.event_type])

        # Trigger processing
        process_event(self.event)

        # Event is captured and POST request is triggered because event_type and user_uuid match
        requests_post.assert_called_once_with(
            self.web_hook.destination_url, json=mock.ANY, verify=settings.VERIFY_WEBHOOK_REQUESTS)

    def test_email_hook_processor_can_be_called_twice(self):
        # Create email hook for customer owner
        email_hook = logging_models.EmailHook.objects.create(user=self.owner,
                                                             email=self.owner.email,
                                                             event_types=[self.event_type])

        # If event is not mutated, exception is not raised, see also SENTRY-1396
        email_hook.process(self.event)
        email_hook.process(self.event)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0031_add_options_default'),
    ]

    operations = [
        migrations.AlterField(
            model_name='servicesettings',
            name='options',
            field=waldur_core.core.fields.JSONField(default={}, help_text='Extra options', blank=True),
            preserve_default=True,
        ),
    ]



from __future__ import unicode_literals

from django.conf import settings as django_settings
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _
import django_filters
from django_filters.widgets import BooleanWidget
from rest_framework import settings, filters
from rest_framework.serializers import ValidationError

from waldur_core.core import serializers as core_serializers, filters as core_filters
from waldur_core.core.filters import ExternalFilterBackend
from waldur_core.core.utils import camel_case_to_underscore
from waldur_core.logging import models, utils
from waldur_core.logging.elasticsearch_client import EmptyQueryset
from waldur_core.logging.loggers import event_logger, expand_event_groups, expand_alert_groups


def format_raw_field(key):
    """
    When ElasticSearch analyzes string, it breaks it into parts.
    In order make query for not-analyzed exact string values, we should use subfield instead.

    The index template for Elasticsearch 5.0 has been changed.
    The subfield for string multi-fields has changed from .raw to .keyword

    Thus workaround for backward compatibility during migration is required.
    See also: https://github.com/elastic/logstash/blob/v5.4.1/docs/static/breaking-changes.asciidoc
    """
    subfield = django_settings.WALDUR_CORE.get('ELASTICSEARCH', {}).get('raw_subfield', 'keyword')
    return '%s.%s' % (camel_case_to_underscore(key), subfield)


class EventFilterBackend(filters.BaseFilterBackend):
    """ Sorting is supported in ascending and descending order by specifying a field to
        an **?o=** parameter. By default events are sorted by @timestamp in descending order.

        - ?o=\@timestamp

        Filtering of customer list is supported through HTTP query parameters, the following fields are supported:

        - ?event_type=<string> - type of filtered events. Can be list
        - ?search=<string> - text for FTS. FTS fields: 'message', 'customer_abbreviation', 'importance'
          'project_name', 'user_full_name', 'user_native_name'
        - ?scope=<URL> - url of object that is connected to event
        - ?scope_type=<string> - name of scope type of object that is connected to event
        - ?feature=<feature> (can be list) - include all event with type that belong to given features
        - ?exclude_features=<feature> (can be list) - exclude event from output if
          it's type corresponds to one of listed features
        - ?user_username=<string> - user's username
        - ?from=<timestamp> - beginning UNIX timestamp
        - ?to=<timestamp> - ending UNIX timestamp
    """

    def filter_queryset(self, request, queryset, view):
        search_text = request.query_params.get(settings.api_settings.SEARCH_PARAM, '')
        must_terms = {}
        must_not_terms = {}
        should_terms = {}
        excluded_event_types = set()

        if 'event_type' in request.query_params:
            must_terms['event_type'] = request.query_params.getlist('event_type')

        if 'feature' in request.query_params:
            features = request.query_params.getlist('feature')
            must_terms['event_type'] = expand_event_groups(features)

        # Group events by features in order to prevent large HTTP GET request
        if 'exclude_features' in request.query_params:
            features = request.query_params.getlist('exclude_features')
            excluded_event_types.update(expand_event_groups(features))

        if 'exclude_extra' in request.query_params:
            excluded_event_types.update(expand_event_groups(['update']))

        if not django_settings.DEBUG:
            excluded_event_types.update(expand_event_groups(['debug_only']))

        if excluded_event_types:
            must_not_terms['event_type'] = list(excluded_event_types)

        if 'user_username' in request.query_params:
            must_terms['user_username'] = [request.query_params.get('user_username')]

        if 'scope' in request.query_params:
            field = core_serializers.GenericRelatedField(related_models=utils.get_loggable_models())
            field._context = {'request': request}
            obj = field.to_internal_value(request.query_params['scope'])

            # XXX: Ilja - disabling this hack and re-opening a ticket. Additional analysis is required for
            # a proper resolution
            # # XXX: hack to prevent leaking customer events
            # permitted_uuids = [uuid.hex for uuids in
            #                    obj.get_permitted_objects_uuids(request.user).values() for uuid in uuids]
            # if obj.uuid.hex not in permitted_uuids:
            #     raise ValidationError('You do not have permission to view events for scope %s'
            #                           % request.query_params['scope'])

            for key, val in obj.filter_by_logged_object().items():
                must_terms[format_raw_field(key)] = [val]

        elif 'scope_type' in request.query_params:
            choices = utils.get_scope_types_mapping()
            try:
                scope_type = choices[request.query_params['scope_type']]
            except KeyError:
                raise ValidationError(
                    _('Scope type "%(value)s" is not valid. Has to be one from list: %(items)s.') % dict(
                        value=request.query_params['scope_type'],
                        items=', '.join(choices.keys())
                    ))
            else:
                permitted_items = scope_type.get_permitted_objects_uuids(request.user).items()
                if not permitted_items:
                    return EmptyQueryset()
                for field, uuids in permitted_items:
                    must_terms[field] = [uuid.hex for uuid in uuids]

        elif 'resource_type' in request.query_params and 'resource_uuid' in request.query_params:
            # Filter events by resource type and uuid.
            # Please note, that permission checks are skipped,
            # because we can't check permission for deleted resources.
            # Also note, that resource type validation is skipped as well,
            # because resource type name formatting is defined in structure application,
            # but we don't want to create circular dependency between logging and structure apps.
            # This issue could be fixed by switching resource type name formatting to str(model._meta)
            # as it is done for scope_type parameter validation.
            must_terms[format_raw_field('resource_type')] = [request.query_params['resource_type']]
            must_terms[format_raw_field('resource_uuid')] = [request.query_params['resource_uuid']]

        else:
            should_terms.update(event_logger.get_permitted_objects_uuids(request.user))

        mapped = {
            'start': request.query_params.get('from'),
            'end': request.query_params.get('to'),
        }
        timestamp_interval_serializer = core_serializers.TimestampIntervalSerializer(
            data={k: v for k, v in mapped.items() if v})
        timestamp_interval_serializer.is_valid(raise_exception=True)
        filter_data = timestamp_interval_serializer.get_filter_data()

        queryset = queryset.filter(search_text=search_text,
                                   should_terms=should_terms,
                                   must_terms=must_terms,
                                   must_not_terms=must_not_terms,
                                   start=filter_data.get('start'),
                                   end=filter_data.get('end'))

        order_by = request.query_params.get('o', '-@timestamp')
        if order_by:
            queryset = queryset.order_by(order_by)

        return queryset


class AlertFilter(django_filters.FilterSet):
    """ Basic filters for alerts """

    acknowledged = django_filters.BooleanFilter(name='acknowledged', distinct=True, widget=BooleanWidget)
    closed_from = core_filters.TimestampFilter(name='closed', lookup_expr='gte')
    closed_to = core_filters.TimestampFilter(name='closed', lookup_expr='lt')
    created_from = core_filters.TimestampFilter(name='created', lookup_expr='gte')
    created_to = core_filters.TimestampFilter(name='created', lookup_expr='lt')
    content_type = core_filters.ContentTypeFilter()
    message = django_filters.CharFilter(lookup_expr='icontains')

    o = django_filters.OrderingFilter(fields=('severity', 'created'))

    class Meta:
        model = models.Alert
        fields = [
            'acknowledged',
            'closed_from',
            'closed_to',
            'created_from',
            'created_to',
            'content_type',
            'message'
        ]


class AlertScopeFilterBackend(core_filters.GenericKeyFilterBackend):

    def get_related_models(self):
        return utils.get_loggable_models()

    def get_field_name(self):
        return 'scope'


class AdditionalAlertFilterBackend(filters.BaseFilterBackend):
    """
    Additional filters for alerts.

    Support for filters that are related to more than one field or provides unusual query.
    """

    def filter_queryset(self, request, queryset, view):
        mapped = {
            'start': request.query_params.get('from'),
            'end': request.query_params.get('to'),
        }
        timestamp_interval_serializer = core_serializers.TimestampIntervalSerializer(
            data={k: v for k, v in mapped.items() if v})
        timestamp_interval_serializer.is_valid(raise_exception=True)
        filter_data = timestamp_interval_serializer.get_filter_data()
        if 'start' in filter_data:
            queryset = queryset.filter(
                Q(closed__gte=filter_data['start']) | Q(closed__isnull=True))
        if 'end' in filter_data:
            queryset = queryset.filter(created__lte=filter_data['end'])

        if 'opened' in request.query_params:
            queryset = queryset.filter(closed__isnull=True)

        if 'closed' in request.query_params:
            queryset = queryset.filter(closed__isnull=False)

        if 'severity' in request.query_params:
            severity_codes = {v: k for k, v in models.Alert.SeverityChoices.CHOICES}
            severities = [
                severity_codes.get(severity_name) for severity_name in request.query_params.getlist('severity')]
            queryset = queryset.filter(severity__in=severities)

        # XXX: this filter is wrong and deprecated, need to be removed after replacement in Portal
        if 'scope_type' in request.query_params:
            choices = {camel_case_to_underscore(m.__name__): m for m in utils.get_loggable_models()}
            try:
                scope_type = choices[request.query_params['scope_type']]
            except KeyError:
                raise ValidationError(
                    _('Scope type "%(value)s" is not valid. Has to be one from list: %(items)s.') % dict(
                        value=request.query_params['scope_type'],
                        items=', '.join(choices.keys())
                    ))
            else:
                ct = ContentType.objects.get_for_model(scope_type)
                queryset = queryset.filter(content_type=ct)

        if 'alert_type' in request.query_params:
            queryset = queryset.filter(alert_type__in=request.query_params.getlist('alert_type'))

        # Group alerts by features in order to prevent large HTTP GET request
        if 'exclude_features' in request.query_params:
            features = request.query_params.getlist('exclude_features')
            queryset = queryset.exclude(alert_type__in=expand_alert_groups(features))

        return queryset


class ExternalAlertFilterBackend(ExternalFilterBackend):
    pass


class BaseHookFilter(django_filters.FilterSet):
    author_uuid = django_filters.UUIDFilter(name='user__uuid')
    is_active = django_filters.BooleanFilter(widget=BooleanWidget)
    last_published = django_filters.DateTimeFilter()


class WebHookFilter(BaseHookFilter):
    class Meta(object):
        model = models.WebHook
        fields = ('destination_url', 'content_type')


class EmailHookFilter(BaseHookFilter):
    class Meta(object):
        model = models.EmailHook
        fields = ('email',)


class HookSummaryFilterBackend(core_filters.SummaryFilter):
    def get_queryset_filter(self, queryset):
        if queryset.model == models.WebHook:
            return WebHookFilter
        elif queryset.model == models.EmailHook:
            return EmailHookFilter
        elif queryset.model == models.PushHook:
            return PushHookFilter

        return BaseHookFilter

    def get_base_filter(self):
        return BaseHookFilter


class PushHookFilter(BaseHookFilter):
    class Meta(object):
        model = models.PushHook
        fields = ('type', 'device_id', 'device_manufacturer', 'device_model', 'token')



from django.contrib import admin
from django.contrib.contenttypes import models as ct_models
from django.contrib.contenttypes.admin import GenericTabularInline
from django.forms import ModelForm

from waldur_core.core.admin import ReversionAdmin
from waldur_core.quotas import models, utils


class QuotaScopeClassListFilter(admin.SimpleListFilter):
    # Human-readable title
    title = 'Scope class'

    # Parameter for the filter that will be used in the URL query
    parameter_name = 'scope_class'

    def lookups(self, request, model_admin):
        models = utils.get_models_with_quotas()
        return [(ct_models.ContentType.objects.get_for_model(m).id, m.__name__) for m in models]

    def queryset(self, request, queryset):
        content_type_id = self.value()
        if content_type_id:
            return queryset.filter(content_type_id=content_type_id)
        return queryset


class QuotaFieldTypeLimit(object):
    fields = ('name', 'limit', 'usage', 'quota_field_type')
    readonly_fields = ('name', 'usage', 'quota_field_type')

    def quota_field_type(self, obj):
        field = obj.get_field()
        if field:
            return field.__class__.__name__
        return ''


class QuotaForm(ModelForm):
    class Meta:
        model = models.Quota
        fields = ('name', 'limit', 'usage')

    # TODO: Temporarily disabled because it breaks validation
    # def __init__(self, *args, **kwargs):
    #     super(QuotaForm, self).__init__(*args, **kwargs)
    #
    #     if (self.instance
    #             and self._is_backend_quota_field(self.instance)
    #             and not settings.WALDUR_CORE['BACKEND_FIELDS_EDITABLE']):
    #         self.fields['limit'].widget = ReadonlyTextWidget()
    #
    # def _is_backend_quota_field(self, quota):
    #     if not quota.scope:
    #         return False
    #
    #     field = getattr(quota.scope.Quotas, quota.name)
    #     return field.is_backend


class QuotaAdmin(QuotaFieldTypeLimit, ReversionAdmin):
    list_display = ['scope', 'name', 'limit', 'usage']
    list_filter = ['name', QuotaScopeClassListFilter]


class QuotaInline(QuotaFieldTypeLimit, GenericTabularInline):
    model = models.Quota
    form = QuotaForm
    extra = 0
    can_delete = False


admin.site.register(models.Quota, QuotaAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-09-20 09:45
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cost_tracking', '0025_increase_price_decimal_places'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='priceestimate',
            name='limit',
        ),
        migrations.RemoveField(
            model_name='priceestimate',
            name='threshold',
        ),
    ]



from django.test import TestCase
from six.moves import mock

from .. import factories, fixtures

from waldur_core.structure import models as structure_models


class LogProjectSaveTest(TestCase):

    @mock.patch('waldur_core.structure.handlers.event_logger')
    def test_logger_called_once_on_project_create(self, logger_mock):
        new_project = factories.ProjectFactory()

        logger_mock.project.info.assert_called_once_with(
            'Project {project_name} has been created.',
            event_type='project_creation_succeeded',
            event_context={
                'project': new_project,
            },
        )

    def test_logger_called_once_on_project_name_update(self):
        new_project = factories.ProjectFactory()
        old_name = new_project.name

        with mock.patch('waldur_core.structure.handlers.event_logger') as logger_mock:
            new_project.name = 'new name'
            new_project.save()

            logger_mock.project.info.assert_called_once_with(
                "Project {project_name} has been updated. Name has been changed from '%s' to '%s'." % (
                    old_name,
                    new_project.name,
                ),
                event_type='project_update_succeeded',
                event_context={
                    'project': new_project,
                },
            )

    def test_logger_logs_project_name_and_description_when_updated(self):
        new_project = factories.ProjectFactory(description='description', name='name')

        with mock.patch('waldur_core.structure.handlers.event_logger') as logger_mock:
            new_project.name = 'new name'
            new_project.description = 'new description'
            new_project.save()

            expected_message = ('Project {project_name} has been updated.'
                                " Description has been changed from 'description' to 'new description'."
                                " Name has been changed from 'name' to 'new name'.")
            logger_mock.project.info.assert_called_once_with(
                expected_message,
                event_type='project_update_succeeded',
                event_context={
                    'project': new_project,
                },
            )


class LogRoleEventTest(TestCase):

    def test_logger_called_when_customer_role_is_granted(self):
        fixture = fixtures.CustomerFixture()

        owner = fixture.owner
        with mock.patch('waldur_core.structure.handlers.event_logger.customer_role.info') as logger_mock:
            fixture.customer.add_user(fixture.user, structure_models.CustomerRole.OWNER, owner)

            logger_mock.assert_called_once_with(
                'User {affected_user_username} has gained role of {role_name} in customer {customer_name}.',
                event_type='role_granted',
                event_context={
                    'customer': fixture.customer,
                    'user': fixture.owner,
                    'affected_user': fixture.user,
                    'structure_type': 'customer',
                    'role_name': 'Owner',
                },
            )

    def test_logger_called_when_customer_role_is_revoked(self):
        fixture = fixtures.CustomerFixture()
        owner = fixture.owner

        with mock.patch('waldur_core.structure.handlers.event_logger.customer_role.info') as logger_mock:
            fixture.customer.remove_user(owner, structure_models.CustomerRole.OWNER, fixture.staff)

            logger_mock.assert_called_once_with(
                'User {affected_user_username} has lost role of {role_name} in customer {customer_name}.',
                event_type='role_revoked',
                event_context={
                    'customer': fixture.customer,
                    'user': fixture.staff,
                    'affected_user': fixture.owner,
                    'structure_type': 'customer',
                    'role_name': 'Owner',
                },
            )

    def test_logger_called_when_project_role_is_granted(self):
        fixture = fixtures.ProjectFixture()

        with mock.patch('waldur_core.structure.handlers.event_logger.project_role.info') as logger_mock:
            fixture.project.add_user(fixture.user, structure_models.ProjectRole.MANAGER, fixture.owner)

            logger_mock.assert_called_once_with(
                'User {affected_user_username} has gained role of {role_name} in project {project_name}.',
                event_type='role_granted',
                event_context={
                    'project': fixture.project,
                    'user': fixture.owner,
                    'affected_user': fixture.user,
                    'structure_type': 'project',
                    'role_name': 'Manager',
                },
            )

    def test_logger_called_when_project_role_is_revoked(self):
        fixture = fixtures.ProjectFixture()
        manager = fixture.manager

        with mock.patch('waldur_core.structure.handlers.event_logger.project_role.info') as logger_mock:
            fixture.project.remove_user(manager, structure_models.ProjectRole.MANAGER, fixture.owner)

            logger_mock.assert_called_once_with(
                'User {affected_user_username} has revoked role of {role_name} in project {project_name}.',
                event_type='role_revoked',
                event_context={
                    'project': fixture.project,
                    'user': fixture.owner,
                    'affected_user': fixture.manager,
                    'structure_type': 'project',
                    'role_name': 'Manager',
                },
            )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django_fsm


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0022_init_global_count_quotas'),
    ]

    operations = [
        migrations.AlterField(
            model_name='servicesettings',
            name='state',
            field=django_fsm.FSMIntegerField(default=5, choices=[(0, 'New'), (5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Sync Scheduled'), (2, 'Syncing'), (3, 'In Sync'), (4, 'Erred')]),
            preserve_default=True,
        ),
    ]



from django.core.management import call_command
from django.test import TestCase
from six import StringIO


class CommandsTestCase(TestCase):
    def test_no_missing_migrations(self):

        result = StringIO()
        call_command('makemigrations', dry_run=True, stdout=result)
        result_string = result.getvalue()
        self.assertEqual(result_string, 'No changes detected\n')



from django.contrib.contenttypes import models as ct_models
from django.db import models
from django.db.models import Q


# XXX: This manager are very similar with quotas manager
class AlertManager(models.Manager):

    def filtered_for_user(self, user, queryset=None):
        from waldur_core.logging import utils

        if queryset is None:
            queryset = self.get_queryset()
        # XXX: This circular dependency will be removed then filter_queryset_for_user
        # will be moved to model manager method
        from waldur_core.structure.managers import filter_queryset_for_user

        query = Q()
        for model in utils.get_loggable_models():
            user_object_ids = filter_queryset_for_user(model.objects.all(), user).values_list('id', flat=True)
            content_type_id = ct_models.ContentType.objects.get_for_model(model).id
            query |= Q(object_id__in=user_object_ids, content_type_id=content_type_id)

        return queryset.filter(query)

    def for_objects(self, qs):
        kwargs = dict(
            content_type=ct_models.ContentType.objects.get_for_model(qs.model),
            object_id__in=qs.values_list('id', flat=True),
            closed__isnull=True
        )
        return self.get_queryset().filter(**kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0037_remove_customer_billing_backend_id'),
    ]

    operations = [
        migrations.CreateModel(
            name='Invitation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('state', models.CharField(default='pending', max_length=8, choices=[('accepted', 'Accepted'), ('canceled', 'Canceled'), ('pending', 'Pending'), ('expired', 'Expired')])),
                ('link_template', models.CharField(help_text='The template must include {uuid} parameter e.g. http://example.com/invitation/{uuid}', max_length=255)),
                ('email', models.EmailField(help_text='Invitation link will be sent to this email. Note that user can accept invitation with different email.', max_length=254)),
                ('customer', models.ForeignKey(verbose_name='organization', related_name='invitations', to='structure.Customer')),
                ('customer_role', models.ForeignKey(related_name='invitations', blank=True, to='structure.CustomerRole', null=True)),
                ('project_role', models.ForeignKey(related_name='invitations', blank=True, to='structure.ProjectRole', null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from __future__ import unicode_literals

from functools import reduce
import importlib
import inspect
import logging

from django.apps import apps
from django.conf import settings
from django.contrib.admindocs.views import simplify_regex
from django.urls import RegexURLResolver, RegexURLPattern
from django_filters import ModelMultipleChoiceFilter
from rest_framework.fields import ChoiceField, ReadOnlyField, ModelField
from rest_framework.relations import HyperlinkedRelatedField, ManyRelatedField
from rest_framework.serializers import ListSerializer, ModelSerializer
from rest_framework.settings import api_settings
from rest_framework.views import APIView
import six

from waldur_core.core.filters import ContentTypeFilter, MappedMultipleChoiceFilter
from waldur_core.core.serializers import GenericRelatedField
from waldur_core.core.utils import get_fake_context

logger = logging.getLogger(__name__)


def getdoc(obj, warning=True):
    doc = inspect.getdoc(obj) or ''
    if not doc and warning:
        if inspect.isclass(obj):
            name = '{}.{}'.format(obj.__module__, obj.__name__)
        elif inspect.ismethod(obj):
            cls = obj.im_class
            name = '{}.{}.{}'.format(cls.__module__, cls.__name__, obj.im_func.func_name)
        else:
            name = six.text_type(obj)
        logger.warning("Docstring is missing for %s", name)
    return doc


class ApiDocs(object):
    """ Generate RST docs for DRF endpoints from docstrings:
        - AppConfig class docstring may contain general info about an app,
          `verbose_name` in the config delivers human friendly title;
        - View class docstring describes intention of top-level endpoint;
        - View method docstring can explain usage of particular method or actions;
    """
    tree = {}
    exclude = ['rest_framework']

    def __init__(self, apps=None):
        root_urlconf = importlib.import_module(settings.ROOT_URLCONF)
        endpoints = self.get_all_view_names(root_urlconf.urlpatterns)
        self.build_docs_tree(endpoints)
        self.apps = apps

    def build_docs_tree(self, endpoints):
        for ep in endpoints:
            if ep.app in self.exclude:
                continue
            root = '/'.join(ep.path.split('/')[:3])
            if not root.endswith('/'):
                root += '/'
            self.tree.setdefault(ep.app, {})
            self.tree[ep.app].setdefault(root, [])
            self.tree[ep.app][root].append(ep)

    def get_all_view_names(self, urlpatterns, parent_pattern=None):
        for pattern in urlpatterns:
            if isinstance(pattern, RegexURLResolver):
                pp = None if pattern._regex == '^' else pattern
                for ep in self.get_all_view_names(pattern.url_patterns, parent_pattern=pp):
                    yield ep
            elif isinstance(pattern, RegexURLPattern) and self._is_drf_view(pattern):
                suffix = '?P<%s>' % api_settings.FORMAT_SUFFIX_KWARG
                if suffix not in pattern.regex.pattern:
                    yield ApiEndpoint(pattern, parent_pattern)

    def _is_drf_view(self, pattern):
        return hasattr(pattern.callback, 'cls') and issubclass(pattern.callback.cls, APIView)

    def _get_fields(self, fields):
        lines = []
        for field in fields:
            hl = '**' if field['required'] else ''
            cmnt = ' (%s)' % field['help_text'] if field['help_text'] else ''
            txt = '\t* {hl}{name}{hl} -- ``{type}``{cmnt}'.format(hl=hl, cmnt=cmnt, **field)
            lines.append(txt)
        return '\n'.join(lines)

    def generate(self, path):
        for app, endpoints in sorted(self.tree.items()):
            if self.apps and app not in self.apps:
                continue

            conf = apps.get_app_config(app)
            name = conf.verbose_name
            file = '%s.rst' % app

            print '\t* %s' % file

            with open(path + '/' + file, 'w') as f:
                doc = getdoc(conf) or name
                f.write(name + '\n' + '=' * len(name) + '\n\n')
                f.write(doc + '\n\n')
                for endpoint, actions in sorted(endpoints.items(), reverse=True):
                    top = actions[0]
                    fields = top.get_serializer_fields()

                    f.write(endpoint + '\n' + '-' * len(endpoint) + '\n\n')
                    if top.docstring:
                        f.write(top.docstring + '\n\n')

                    doc = top.get_filter_docs()
                    if doc:
                        f.write(doc + '\n')

                    f.write('Supported actions and methods:\n\n')
                    for idx, act in enumerate(actions, start=1):
                        methods = act.methods
                        # 1st line is supposed to be List/Create view -- remove unfeasible methods
                        if idx == 1:
                            methods = [m for m in methods if m not in ('PUT', 'PATCH', 'DELETE')]
                        # 2nd line is supposed to be Retrieve/Update/Delete view
                        if idx == 2 and act.path.endswith('>/'):
                            methods = [m for m in methods if m != 'POST']

                        f.write('.. topic:: ``%s``' % act.path + '\n\n')
                        f.write('\tMethods: ' + ', '.join(['``%s``' % m for m in methods]) + '\n\n')

                        # 1st line is supposed to be List/Create view -- add proper details
                        if idx == 1:
                            create_fields = [o for o in fields if not o['readonly']]
                            if 'POST' in methods and create_fields:
                                f.write('\tSupported fields for creation:\n\n')
                                f.write(self._get_fields(create_fields))
                                f.write('\n')

                        # 2nd line is supposed to be Retrieve/Update/Delete view
                        if idx == 2 and act.path.endswith('>/'):
                            update_fields = [o for o in fields if not o['readonly'] and not o['protected']]
                            if 'PUT' in methods and update_fields:
                                f.write('\tSupported fields for update:\n\n')
                                f.write(self._get_fields(update_fields))
                                f.write('\n')

                        cls = act.callback.cls
                        if act.docstring:
                            self._write_docstring(f, act.docstring)

                        if act.action:
                            doc = getdoc(getattr(cls, act.action))
                            if doc:
                                self._write_docstring(f, doc)
                        else:
                            # docs for ViewSets
                            for method, a in act.METHODS:
                                if idx == 1 and a == 'retrieve':
                                    continue
                                if idx >= 2 and a == 'list':
                                    continue

                                action = getattr(cls, a, None)
                                if action:
                                    doc = getdoc(action, warning=False)
                                    if doc:
                                        self._write_docstring(f, doc)

                            else:
                                continue

                            # docs for classic Views
                            for method in methods:
                                try:
                                    doc = getdoc(getattr(cls, method.lower()), warning=False)
                                except AttributeError:
                                    continue
                                if doc:
                                    self._write_docstring(f, doc)

                    f.write('\n\n')

        with open(path + '/index.rst', 'w') as f:
            lines = [
                "",
                ".. toctree::",
                "   :glob:",
                "   :titlesonly:",
                "   :maxdepth: 1",
                "",
                "   **",
            ]
            f.writelines([l + '\n' for l in lines])

    def _write_docstring(self, file, docstring):
        file.write('\n'.join(['\t' + s for s in docstring.split('\n')]) + '\n')


class ApiEndpoint(object):
    FIELDS = {
        # filter
        'BooleanFilter': 'boolean',
        'CharFilter': 'string',
        'TimestampFilter': 'UNIX timestamp',
        'QuotaFilter': 'float',
        'URLFilter': 'link',
        # serializer
        'BooleanField': 'boolean',
        'CharField': 'string',
        'DecimalField': 'float',
        'FloatField': 'float',
        'FileField': 'file',
        'EmailField': 'email',
        'IntegerField': 'integer',
        'IPAddressField': 'IP address',
        'HyperlinkedRelatedField': 'link',
        'URLField': 'URL',
    }

    METHODS = [
        ('GET', 'list'),
        ('GET', 'retrieve'),
        ('POST', 'create'),
        ('PUT', 'update'),
        ('PATCH', 'partial_update'),
        ('DELETE', 'destroy')
    ]

    VIEWS = {}

    def __init__(self, pattern, parent_pattern=None):
        # XXX: Hotfix for openstack app docs.
        app_name = pattern.callback.cls.__module__.split('.')[-2]
        app_name = app_name.replace('waldur_openstack', 'openstack')
        app_name = app_name.replace('waldur_auth_social', 'nodeconductor_auth')
        conf = apps.get_app_config(app_name)
        self.pattern = pattern
        self.callback = pattern.callback
        self.docstring = getdoc(self.callback.cls)
        self.name_parent = simplify_regex(parent_pattern.regex.pattern).replace('/', '') if parent_pattern else None
        self.path = self.get_path(parent_pattern)
        self.name = conf.verbose_name
        self.app = conf.label

        cls = self.callback.cls
        action = self.path.split('/')[-2]

        self.actions = [m for m in dir(cls) if hasattr(getattr(cls, m), 'bind_to_methods')]
        self.action = action if action in self.actions else None
        if self.action:
            self.methods = [m.upper() for m in reduce(getattr, [action, 'bind_to_methods'], cls)]
        else:
            self.methods = []
            for method, action in self.METHODS:
                if method in self.methods:
                    continue
                if hasattr(cls, action) or hasattr(cls, method.lower()):
                    self.methods.append(method)

        self.VIEWS[pattern.name] = self.path

    def get_serializer_fields(self, cls=None):
        if not cls:
            try:
                cls = self.callback.cls.serializer_class
            except AttributeError:
                return []

        if isinstance(cls, type):
            serializer = cls(context=get_fake_context())
        else:
            serializer = cls
        if cls:
            meta = getattr(serializer, 'Meta', object())
            ro = getattr(meta, 'read_only_fields', [])
            wo = getattr(meta, 'write_only_fields', [])
            pt = getattr(meta, 'protected_fields', [])
            return [{
                "name": key,
                "type": self._get_field_type(field),
                "help_text": field.help_text,
                "required": field.required,
                "readonly": key in ro or isinstance(field, ReadOnlyField) or field.read_only,
                "writeonly": key in wo or field.write_only,
                "protected": key in pt,
            } for key, field in serializer.get_fields().items()]

        return []

    def get_filter_docs(self):
        try:
            docs = []
            for cls in self.callback.cls.filter_backends:
                doc = getdoc(cls)
                if not doc.startswith('..drfdocs-ignore'):
                    docs.append(doc)
            return '\n\n'.join(docs)
        except AttributeError:
            return ''

    def get_path(self, parent_pattern):
        if parent_pattern:
            return "/{}{}".format(self.name_parent, simplify_regex(self.pattern.regex.pattern))
        return simplify_regex(self.pattern.regex.pattern)

    def _get_field_type(self, field):
        if isinstance(field, MappedMultipleChoiceFilter):
            return 'choice(%s)' % ', '.join(["'%s'" % f for f in sorted(field.mapped_to_model)])
        if isinstance(field, ChoiceField):
            return 'choice(%s)' % ', '.join(["'%s'" % f for f in sorted(field.choices)])
        if isinstance(field, HyperlinkedRelatedField):
            path = self.VIEWS.get(field.view_name)
            if path:
                return 'link to %s' % path
        if isinstance(field, GenericRelatedField):
            paths = [self.VIEWS.get(GenericRelatedField()._get_url(m())) for m in field.related_models]
            path = ', '.join([m for m in paths if m])
            if path:
                return 'link to any: %s' % path
        if isinstance(field, ContentTypeFilter):
            return 'string in form <app_label>.<model_name>'
        if isinstance(field, ModelSerializer):
            fields = {f['name']: f['type'] for f in self.get_serializer_fields(field) if not f['readonly']}
            return '{%s}' % ', '.join(['%s: %s' % (k, v) for k, v in fields.items()])
        if isinstance(field, ModelMultipleChoiceFilter):
            return self._get_field_type(field.field)
        if isinstance(field, ListSerializer):
            return 'list of [%s]' % self._get_field_type(field.child)
        if isinstance(field, ManyRelatedField):
            return 'list of [%s]' % self._get_field_type(field.child_relation)
        if isinstance(field, ModelField):
            return self._get_field_type(field.model_field)
        name = field.__class__.__name__
        return self.FIELDS.get(name, name)

    def __repr__(self):
        return "%s -- %s" % (self.path, ', '.join(self.methods))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0039_remove_permission_groups'),
    ]

    operations = [
        migrations.AlterField(
            model_name='customerpermission',
            name='is_active',
            field=models.NullBooleanField(default=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='projectpermission',
            name='is_active',
            field=models.NullBooleanField(default=True, db_index=True),
        ),
    ]



""" Custom loggers that allows to store logs in DB and Elastic """

from __future__ import unicode_literals

from collections import defaultdict
import datetime
import decimal
import importlib
import logging
import types
import uuid

from django.apps import apps
from django.contrib.contenttypes import models as ct_models
from django.db import transaction, IntegrityError
import six

from waldur_core.logging import models
from waldur_core.logging.log import EventLoggerAdapter
from waldur_core.logging.middleware import get_event_context

logger = logging.getLogger(__name__)


class LoggerError(AttributeError):
    pass


class EventLoggerError(AttributeError):
    pass


class AlertLoggerError(AttributeError):
    pass


class BaseLogger(object):
    def __init__(self, logger_name=__name__):
        self._meta = getattr(self, 'Meta', None)
        self.supported_types = self.get_supported_types()

    def get_supported_types(self):
        raise NotImplementedError

    def get_supported_groups(self):
        raise NotImplementedError

    def get_nullable_fields(self):
        return getattr(self._meta, 'nullable_fields', [])

    def get_field_model(self, model):
        if not isinstance(model, six.string_types):
            return model

        try:
            return apps.get_model(model)
        except LookupError:
            try:
                app_name, class_name = model.split('.')
                module = importlib.import_module('waldur_core.%s.models' % app_name)
                return getattr(module, class_name)
            except (ImportError, AttributeError, IndexError):
                raise LoggerError("Can't find model %s" % model)

    def compile_message(self, message_template, context):
        try:
            msg = six.text_type(message_template).format(**context)
        except KeyError as e:
            raise LoggerError(
                "Cannot find %s context field. Choices are: %s" % (
                    six.text_type(e), ', '.join(context.keys())))
        return msg

    def validate_logging_type(self, logging_type):
        if self.supported_types and logging_type not in self.supported_types:
            raise EventLoggerError(
                "Unsupported logging type '%s'. Choices are: %s" % (
                    logging_type, ', '.join(self.supported_types)))

    def compile_context(self, **kwargs):
        # Get a list of fields here in order to be sure all models already loaded.
        if not hasattr(self, 'fields'):
            self.fields = {
                k: self.get_field_model(v)
                for k, v in self.__class__.__dict__.items()
                if not k.startswith('_') and not isinstance(v, types.FunctionType) and k != 'Meta'}

        missed = set(self.fields.keys()) - set(self.get_nullable_fields()) - set(kwargs.keys())
        if missed:
            raise LoggerError("Missed fields in event context: %s" % ', '.join(missed))

        context = {}

        event_context = get_event_context()
        if event_context:
            context.update(event_context)
            username = event_context.get('user_username')
            if 'user' in self.fields and username:
                logger.warning("User is passed directly to event context. "
                               "Currently authenticated user %s is ignored.", username)

        for entity_name, entity in six.iteritems(kwargs):
            if entity_name in self.fields:
                entity_class = self.fields[entity_name]
                if entity is None and entity_name in self.get_nullable_fields():
                    continue
                if not isinstance(entity, entity_class):
                    raise LoggerError(
                        "Field '%s' must be an instance of %s but %s received" % (
                            entity_name, entity_class.__name__, entity.__class__.__name__))
            else:
                logger.error(
                    "Field '%s' cannot be used in logging context for %s",
                    entity_name, self.__class__.__name__)
                continue

            if isinstance(entity, LoggableMixin):
                context.update(entity._get_log_context(entity_name))
            elif isinstance(entity, (int, float, six.string_types, dict, tuple, list, bool)):
                context[entity_name] = entity
            elif entity is None:
                pass
            else:
                context[entity_name] = six.text_type(entity)
                logger.warning(
                    "Cannot properly serialize '%s' context field. "
                    "Must be inherited from LoggableMixin." % entity_name)

        return context


class EventLogger(BaseLogger):
    """ Base event logger API.
        Fields which must be passed during event log emitting (event context)
        should be defined as attributes for this class in the form of:

        field_name = ObjectClass || '<app_label>.<class_name>'

        A list of supported event types can be defined with help of method get_supported_types,
        or 'event_types' property of Meta class. Event type won't be validated if this list is empty.

        Example usage:

        .. code-block:: python

            from waldur_core.openstack.models import Tenant

            class QuotaEventLogger(EventLogger):
                tenant = Tenant
                project = 'structure.Project'
                threshold = float
                quota_type = six.string_types

                class Meta:
                    event_types = 'quota_threshold_reached',


            quota_logger = QuotaEventLogger(__name__)
            quota_logger.warning(
                '{quota_type} quota threshold has been reached for {project_name}.',
                event_type='quota_threshold_reached',
                event_context=dict(
                    quota_type=quota.name,
                    project=membership.project,
                    tenant=tenant,
                    threshold=threshold * quota.limit)
            )
    """

    def __init__(self, logger_name=__name__):
        super(EventLogger, self).__init__(logger_name)
        self.logger = EventLoggerAdapter(logging.getLogger(logger_name))

    def get_supported_types(self):
        return getattr(self._meta, 'event_types', tuple())

    def get_supported_groups(self):
        return getattr(self._meta, 'event_groups', {})

    def info(self, *args, **kwargs):
        self.process('info', *args, **kwargs)

    def error(self, *args, **kwargs):
        self.process('error', *args, **kwargs)

    def warning(self, *args, **kwargs):
        self.process('warning', *args, **kwargs)

    def debug(self, *args, **kwargs):
        self.process('debug', *args, **kwargs)

    def process(self, level, message_template, event_type='undefined', event_context=None):
        self.validate_logging_type(event_type)

        if not event_context:
            event_context = {}

        context = self.compile_context(**event_context)
        msg = self.compile_message(message_template, context)

        log = getattr(self.logger, level)
        log(msg, extra={'event_type': event_type, 'event_context': context})


class AlertLogger(BaseLogger):
    """ Base alert logger API.

        Fields which must be passed during alert log emitting (alert context)
        should be defined as attributes for this class in the form of:

        field_name = ObjectClass || '<app_label>.<class_name>'

        A list of supported event types can be defined with help of method get_supported_types,
        or 'alert_types' property of Meta class. Event type won't be validated if this list is empty.

        Example usage:

        .. code-block:: python

            from waldur_core.logging.loggers import AlertLogger, alert_logger
            from waldur_core.quotas import models

            class QuotaAlertLogger(AlertLogger):
                quota = models.Quota

                class Meta:
                    alert_types = ('quota_usage_is_over_threshold', )

            alert_logger.register('quota', QuotaAlertLogger)


            alert_logger.quota.warning(
                'Quota {quota_name} is over threshold. Limit: {quota_limit}, usage: {quota_usage}',
                scope=quota,
                alert_type='quota_usage_is_over_threshold',
                alert_context={
                    'quota': quota
                })
    """

    def get_supported_types(self):
        return getattr(self._meta, 'alert_types', tuple())

    def get_supported_groups(self):
        return getattr(self._meta, 'alert_groups', {})

    def info(self, *args, **kwargs):
        return self.process(models.Alert.SeverityChoices.INFO, *args, **kwargs)

    def error(self, *args, **kwargs):
        return self.process(models.Alert.SeverityChoices.ERROR, *args, **kwargs)

    def warning(self, *args, **kwargs):
        return self.process(models.Alert.SeverityChoices.WARNING, *args, **kwargs)

    def debug(self, *args, **kwargs):
        return self.process(models.Alert.SeverityChoices.DEBUG, *args, **kwargs)

    def process(self, severity, message_template, scope, alert_type='undefined', alert_context=None, fail_silently=True):
        self.validate_logging_type(alert_type)

        if not alert_context:
            alert_context = {}

        context = self.compile_context(**alert_context)
        msg = self.compile_message(message_template, context)
        content_type = ct_models.ContentType.objects.get_for_model(scope)

        try:
            with transaction.atomic():
                alert = models.Alert.objects.select_for_update().get(
                    content_type=content_type,
                    object_id=scope.id,
                    alert_type=alert_type,
                    closed__isnull=True
                )
                if alert.severity != severity or alert.message != msg:
                    alert.severity = severity
                    alert.message = msg
                    alert.save()

                    logger.info(
                        'Updated alert for scope %s (id: %s), with type %s',
                        scope, scope.id, alert_type)

                return alert, False
        except models.Alert.DoesNotExist:
            pass

        try:
            alert = models.Alert.objects.create(
                scope=scope,
                alert_type=alert_type,
                severity=severity,
                message=msg,
                context=context
            )
            logger.info(
                'Created new alert for scope %s (id: %s), with type %s',
                scope, scope.id, alert_type)
            return alert, True
        except IntegrityError:
            logger.warning(
                'Could not create alert for scope %s (id: %s), with type %s due to concurrent update',
                scope, scope.id, alert_type)
            if fail_silently:
                return None, False
            else:
                raise

    def close(self, scope, alert_type):
        try:
            content_type = ct_models.ContentType.objects.get_for_model(scope)
            alert = models.Alert.objects.get(
                object_id=scope.id, content_type=content_type, alert_type=alert_type, closed__isnull=True)
            alert.close()
        except models.Alert.DoesNotExist:
            pass


class LoggableMixin(object):
    """ Mixin to serialize model in logs.
        Extends django model or custom class with fields extraction method.
    """

    def get_log_fields(self):
        return ('uuid', 'name')

    def filter_by_logged_object(self):
        """
        Return query dictionary to search current object in ElasticSearch.
        Model may return custom query, but backend doesn't need to know details.
        """
        return {
            self.__class__.__name__ + '_uuid': self.uuid.hex
        }

    def _get_log_context(self, entity_name):

        context = {}
        for field in self.get_log_fields():
            if not hasattr(self, field):
                continue

            value = getattr(self, field)

            name = "{}_{}".format(entity_name, field)
            if isinstance(value, uuid.UUID):
                context[name] = value.hex
            elif isinstance(value, LoggableMixin):
                context.update(value._get_log_context(field))
            elif isinstance(value, datetime.date):
                context[name] = value.isoformat()
            elif isinstance(value, decimal.Decimal):
                context[name] = float(value)
            elif isinstance(value, dict):
                context[name] = value
            else:
                context[name] = six.text_type(value)

        return context

    @classmethod
    def get_permitted_objects_uuids(cls, user):
        return {}


class BaseLoggerRegistry(object):

    def get_loggers(self):
        raise NotImplementedError('Method "get_loggers" is not implemented.')

    def register(self, name, logger_class):
        if name in self.__dict__:
            raise EventLoggerError("Logger '%s' already registered." % name)
        self.__dict__[name] = logger_class()

    def unregister_all(self):
        self.__dict__ = {}

    def get_all_types(self):
        events = set()
        for elogger in self.get_loggers():
            events.update(elogger.get_supported_types())
        return list(sorted(events))

    def get_all_groups(self):
        groups = defaultdict(set)
        for elogger in self.get_loggers():
            for group, items in elogger.get_supported_groups().items():
                groups[group].update(set(items))
        return dict(groups)

    def expand_groups(self, groups):
        items = set()
        mapping = self.get_all_groups()
        for group in groups:
            items.update(mapping.get(group, []))
        return sorted(items)


class EventLoggerRegistry(BaseLoggerRegistry):

    def get_loggers(self):
        return [l for l in self.__dict__.values() if isinstance(l, EventLogger)]

    def get_permitted_objects_uuids(self, user):
        from waldur_core.logging.utils import get_loggable_models
        permitted_objects_uuids = {}
        for model in get_loggable_models():
            for field, uuids in model.get_permitted_objects_uuids(user).items():
                permitted_objects_uuids[field] = [uuid_obj.hex for uuid_obj in uuids]
        return permitted_objects_uuids


class AlertLoggerRegistry(BaseLoggerRegistry):

    def get_loggers(self):
        return [l for l in self.__dict__.values() if isinstance(l, AlertLogger)]


def get_valid_events():
    return event_logger.get_all_types()


def get_valid_alerts():
    return alert_logger.get_all_types()


def get_event_groups():
    return event_logger.get_all_groups()


def get_event_groups_keys():
    return sorted(get_event_groups().keys())


def get_alert_groups():
    return alert_logger.get_all_groups()


def expand_event_groups(groups):
    return event_logger.expand_groups(groups)


def expand_alert_groups(groups):
    return alert_logger.expand_groups(groups)


class CustomEventLogger(EventLogger):
    """
    Logger for custom events e.g. created by staff user
    """
    scope = LoggableMixin

    class Meta:
        event_types = ('custom_notification',)
        nullable_fields = ('scope',)


class ThresholdAlertLogger(AlertLogger):
    object = 'logging.AlertThresholdMixin'

    class Meta:
        alert_types = ('threshold_exceeded',)


# This global objects represent the default loggers registry
event_logger = EventLoggerRegistry()
alert_logger = AlertLoggerRegistry()

event_logger.register('custom', CustomEventLogger)
alert_logger.register('threshold', ThresholdAlertLogger)



from django.contrib import auth
from rest_framework import serializers, test
from six.moves import mock

from .. import exceptions, mixins

User = auth.get_user_model()


class TestUserSerializer(serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = User
        fields = ('username', 'full_name')


class ExecutorMixinTest(test.APITransactionTestCase):
    def setUp(self):
        self.executor = mixins.CreateExecutorMixin()
        self.executor.create_executor = mock.Mock()

        self.serializer = TestUserSerializer(data={
            'username': 'alice2017',
            'full_name': 'Alice Lebowski',
        })
        self.serializer.is_valid()

    def test_if_executor_succeeds_database_object_is_saved(self):
        self.executor.perform_create(self.serializer)
        self.assertTrue(User.objects.filter(username='alice2017').exists())

    def test_if_executor_fails_and_atomic_transaction_is_not_used_database_object_is_saved(self):
        self.executor.use_atomic_transaction = False
        self.executor.create_executor.execute.side_effect = exceptions.IncorrectStateException()
        self.assertRaises(exceptions.IncorrectStateException, self.executor.perform_create, self.serializer)
        self.assertTrue(User.objects.filter(username='alice2017').exists())

    def test_if_executor_fails_and_atomic_transaction_is_used_database_object_is_not_saved(self):
        self.executor.use_atomic_transaction = True
        self.executor.create_executor.execute.side_effect = exceptions.IncorrectStateException()
        self.assertRaises(exceptions.IncorrectStateException, self.executor.perform_create, self.serializer)
        self.assertFalse(User.objects.filter(username='alice2017').exists())



from django.db import models

from waldur_core.quotas.fields import QuotaField
from waldur_core.quotas.models import QuotaModelMixin
from waldur_core.structure import models as structure_models


class TestService(structure_models.Service):
    projects = models.ManyToManyField(structure_models.Project, through='TestServiceProjectLink')

    @classmethod
    def get_url_name(cls):
        return 'test'


class TestServiceProjectLink(structure_models.ServiceProjectLink):
    service = models.ForeignKey(TestService)

    class Quotas(QuotaModelMixin.Quotas):
        vcpu = QuotaField(default_limit=20, is_backend=True)
        ram = QuotaField(default_limit=51200, is_backend=True)
        storage = QuotaField(default_limit=1024000, is_backend=True)
        instances = QuotaField(default_limit=30, is_backend=True)
        security_group_count = QuotaField(default_limit=100, is_backend=True)
        security_group_rule_count = QuotaField(default_limit=100, is_backend=True)
        floating_ip_count = QuotaField(default_limit=50, is_backend=True)

    @classmethod
    def get_url_name(cls):
        return 'test-spl'


class TestNewInstance(QuotaModelMixin, structure_models.VirtualMachine):

    service_project_link = models.ForeignKey(TestServiceProjectLink, on_delete=models.PROTECT)
    flavor_name = models.CharField(max_length=255, blank=True)

    class Quotas(QuotaModelMixin.Quotas):
        test_quota = QuotaField(default_limit=1)

    @classmethod
    def get_url_name(cls):
        return 'test-new-instances'

    @property
    def internal_ips(self):
        return ['127.0.0.1']

    @property
    def external_ips(self):
        return ['8.8.8.8']


class TestSubResource(structure_models.SubResource):

    service_project_link = models.ForeignKey(TestServiceProjectLink, on_delete=models.PROTECT)


class TestVolume(structure_models.Volume):

    service_project_link = models.ForeignKey(TestServiceProjectLink, on_delete=models.PROTECT)


class TestSnapshot(structure_models.Snapshot):

    service_project_link = models.ForeignKey(TestServiceProjectLink, on_delete=models.PROTECT)



#!/usr/bin/env python
from setuptools import setup, find_packages

from waldur_core.core.test_runner import Test

dev_requires = [
    'Sphinx==1.2.2',
]

tests_requires = [
    'ddt>=1.0.0,<1.1.0',
    'factory_boy==2.4.1',
    'freezegun==0.3.7',
    'mock>=1.0.1',
    'mock-django==0.6.9',
    'six>=1.9.0',
    'sqlalchemy>=1.0.12',
]

install_requires = [
    'Celery==4.1.0',
    'croniter>=0.3.4,<0.3.6',
    'cryptography',
    'Django>=1.11,<2.0',
    'django-admin-tools==0.8.0',
    'django-defender>=0.5.3',
    'django-filter==1.0.2',
    'django-fluent-dashboard==0.6.1',
    'django-fsm==2.3.0',
    'django-jsoneditor>=0.0.7',
    'django-model-utils==3.0.0',
    'django-redis-cache>=1.6.5',
    'django-reversion==2.0.8',
    'django-rest-swagger==2.1.2',
    'django-taggit>=0.20.2',
    'djangorestframework>=3.6.3,<3.7.0',
    'elasticsearch==5.4.0',
    'hiredis>=0.2.0',
    'iptools>=0.6.1',
    'PrettyTable<0.8,>=0.7.1',
    'Pillow>=2.0.0',
    'psycopg2>=2.5.4',  # https://docs.djangoproject.com/en/1.11/ref/databases/#postgresql-notes
    'PyYAML>=3.10',
    'pycountry>=1.20,<2.0',
    'pyvat>=1.3.1,<2.0',
    'redis==2.10.6',
    'requests>=2.6.0,!=2.12.2,!=2.13.0',
    'sqlparse>=0.1.11',
]

setup(
    name='waldur_core',
    version='0.160.3',
    author='OpenNode Team',
    author_email='info@opennodecloud.com',
    url='https://github.com/opennode/waldur-core',
    description='Waldur core is a core of Waldur cloud management service.',
    long_description=open('README.rst').read(),
    packages=find_packages(),
    install_requires=install_requires,
    extras_require={
        'dev': dev_requires,
        'tests': tests_requires,
    },
    entry_points={
        'console_scripts': (
            'waldur = waldur_core.server.manage:main',
        ),
    },
    tests_require=tests_requires,
    cmdclass={'test': Test},
    include_package_data=True,
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import random

from django.test import TestCase

from ..models import GrandparentModel
from ... import exceptions


class QuotaModelMixinTest(TestCase):

    def test_default_quota_is_unlimited(self):
        instance = GrandparentModel.objects.create()
        self.assertEqual(instance.quotas.get(name='regular_quota').limit, -1)

    def test_quota_with_default_limit(self):
        instance = GrandparentModel.objects.create()
        self.assertEqual(instance.quotas.get(name='quota_with_default_limit').limit, 100)

    def test_add_usage_validates_with_unlimited_quota(self):
        instance = GrandparentModel.objects.create()
        try:
            instance.add_quota_usage('regular_quota', 10, validate=True)
        except exceptions.QuotaValidationError:
            self.fail('add_quota_usage should not raise exception if quota is unlimited')

    def test_add_usage_skips_validation_with_limited_quota_but_negative_delta(self):
        instance = GrandparentModel.objects.create()
        try:
            instance.add_quota_usage('quota_with_default_limit', -10, validate=True)
        except exceptions.QuotaValidationError:
            self.fail('add_quota_usage should not raise exception if delta is negative')

    def test_add_usage_fails_if_quota_is_over_limit(self):
        instance = GrandparentModel.objects.create()
        self.assertRaises(exceptions.QuotaValidationError,
                          instance.add_quota_usage,
                          quota_name='quota_with_default_limit',
                          usage_delta=200,
                          validate=True)

    def test_quotas_sum_calculation_if_all_values_are_positive(self):
        # we have 3 memberships:
        instances = [GrandparentModel.objects.create() for _ in range(3)]

        # each membership has non zero quotas:
        for instance in instances:
            for quota_name in instance.get_quotas_names():
                limit = random.choice([10, 20, 30, 40])
                instance.set_quota_limit(quota_name, limit)
                instance.set_quota_usage(quota_name, limit / 2)
        owners = instances[:2]

        sum_of_quotas = GrandparentModel.get_sum_of_quotas_as_dict(owners)

        expected_sum_of_quotas = {}
        for quota_name in GrandparentModel.get_quotas_names():
            expected_sum_of_quotas[quota_name] = sum(owner.quotas.get(name=quota_name).limit for owner in owners)
            expected_sum_of_quotas[quota_name + '_usage'] = sum(
                owner.quotas.get(name=quota_name).usage for owner in owners)

        self.assertEqual(expected_sum_of_quotas, sum_of_quotas)

    def test_quotas_sum_calculation_if_some_limit_is_negative(self):
        instances = [GrandparentModel.objects.create() for _ in range(3)]
        instances[0].set_quota_limit('regular_quota', -1)
        instances[1].set_quota_limit('regular_quota', 10)
        instances[2].set_quota_limit('regular_quota', 30)

        sum_of_quotas = GrandparentModel.get_sum_of_quotas_as_dict(
            instances, quota_names=['regular_quota'], fields=['limit'])
        self.assertEqual({'regular_quota': -1}, sum_of_quotas)

    def test_quotas_sum_calculation_if_all_limits_are_negative(self):
        instances = [GrandparentModel.objects.create() for _ in range(3)]
        instances[0].set_quota_limit('regular_quota', -1)
        instances[1].set_quota_limit('regular_quota', -1)
        instances[2].set_quota_limit('regular_quota', -1)

        sum_of_quotas = GrandparentModel.get_sum_of_quotas_as_dict(
            instances, quota_names=['regular_quota'], fields=['limit'])
        self.assertEqual({'regular_quota': -1}, sum_of_quotas)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0025_add_zabbix_to_settings'),
    ]

    operations = [
        migrations.AddField(
            model_name='servicesettings',
            name='error_message',
            field=models.TextField(blank=True),
            preserve_default=True,
        ),
    ]



from django.test import TestCase
from six.moves import mock

from waldur_core.core import views


class TestPublicSettings(TestCase):

    def setUp(self):
        super(TestPublicSettings, self).setUp()

        class MockExtension(object):
            def __init__(self, name):
                class Settings(object):
                    def __init__(self, name):
                        setattr(self, name, {})

                self.Settings = Settings(name)

            @staticmethod
            def get_public_settings():
                return ['INFO']

        extensions = {
            'WALDUR_EXTENSION_1': {'ENABLED': False},
            'WALDUR_EXTENSION_2': {'ENABLED': True},
            'WALDUR_EXTENSION_3': {'SECRET': 'secret', 'INFO': 'info'}
        }
        mock_settings = mock.Mock(WALDUR_CORE={}, WALDUR_CORE_PUBLIC_SETTINGS=[], **extensions)
        self.patcher_settings = mock.patch('waldur_core.core.views.settings', new=mock_settings)
        self.patcher_settings.start()

        self.patcher = mock.patch('waldur_core.core.views.WaldurExtension')
        self.mock = self.patcher.start()
        self.mock.get_extensions.return_value = [MockExtension(e) for e in extensions.keys()]

    def tearDown(self):
        super(TestPublicSettings, self).tearDown()
        mock.patch.stopall()

    def test_if_extension_not_have_field_enabled_or_it_equally_true_this_extension_must_by_in_response(self):
        response = views.get_public_settings()
        self.assertTrue('WALDUR_EXTENSION_2' in response.keys())
        self.assertTrue('WALDUR_EXTENSION_3' in response.keys())

    def test_if_extension_have_field_enabled_and_it_equally_false_this_extension_not_to_be_in_response(self):
        response = views.get_public_settings()
        self.assertFalse('WALDUR_EXTENSION_1' in response.keys())

    def test_if_field_in_get_public_settings_it_value_must_by_in_response(self):
        response = views.get_public_settings()
        self.assertTrue('INFO' in response['WALDUR_EXTENSION_3'])

    def test_if_field_not_in_get_public_settings_it_value_not_to_be_in_response(self):
        response = views.get_public_settings()
        self.assertFalse('SECRET' in response['WALDUR_EXTENSION_3'])



# Django test settings for Waldur Core.
from waldur_core.server.doc_settings import *

MEDIA_ROOT = '/tmp/'

INSTALLED_APPS += (
    'waldur_core.quotas.tests',
    'waldur_core.structure.tests',
)

ROOT_URLCONF = 'waldur_core.structure.tests.urls'



from django.test import TestCase
from rest_framework import serializers

from waldur_core.structure.metadata import ActionsMetadata


class ResourceProvisioningMetadataTest(TestCase):
    def get_serializer(self):
        STATE_CHOICES = (
            (1, 'Ready'),
            (2, 'Erred')
        )

        class Queryset(object):
            def all(self):
                return {'key': 'value'}

            def __iter__(self):
                return iter([])

        class VirtualMachineSerializer(serializers.Serializer):
            name = serializers.CharField(max_length=100, read_only=True)
            description = serializers.CharField(max_length=100)

            state = serializers.ChoiceField(choices=STATE_CHOICES)
            image = serializers.RelatedField(queryset=Queryset())

        return VirtualMachineSerializer()

    def test_read_only_options_are_skipped(self):
        options = ActionsMetadata()

        serializer = self.get_serializer()
        serializer_info = options.get_serializer_info(serializer)

        self.assertIn('description', serializer_info)
        self.assertNotIn('name', serializer_info)

    def test_choices_for_related_fields_are_not_exposed(self):
        options = ActionsMetadata()

        serializer = self.get_serializer()
        serializer_info = options.get_serializer_info(serializer)

        self.assertIn('choices', serializer_info['state'])
        self.assertNotIn('choices', serializer_info['image'])



# coding=utf-8
from __future__ import unicode_literals

from django.core.management import call_command
from django.test import TestCase
import six
from six import StringIO

from .. import factories


class DumpUsersCommandTest(TestCase):

    def test_dump_users_command_with_unicode_full_name(self):
        user = factories.UserFactory(full_name='äöur šipākøv')
        output = StringIO()
        try:
            call_command('dumpusers', stdout=output)
        except UnicodeDecodeError as e:
            self.fail(str(e))
        value = output.getvalue()
        if not isinstance(value, six.text_type):
            value = value.decode('utf-8')
        self.assertIn(user.full_name, value)



import datetime

from django.contrib.contenttypes.models import ContentType
from django.test import TransactionTestCase
from freezegun import freeze_time

from waldur_core.cost_tracking import models, CostTrackingRegister, tasks
from waldur_core.cost_tracking.tests import factories
from waldur_core.structure.tests import factories as structure_factories
from waldur_core.structure.tests.models import TestNewInstance


class RecalculateEstimateTest(TransactionTestCase):

    def setUp(self):
        resource_content_type = ContentType.objects.get_for_model(TestNewInstance)
        self.price_list_item = models.DefaultPriceListItem.objects.create(
            item_type='storage', key='1 MB', resource_content_type=resource_content_type, value=2)
        CostTrackingRegister.register_strategy(factories.TestNewInstanceCostTrackingStrategy)
        self.start_time = datetime.datetime(2016, 8, 8, 11, 0)
        with freeze_time(self.start_time):
            self.resource = structure_factories.TestNewInstanceFactory(disk=20 * 1024)
        self.spl = self.resource.service_project_link
        self.project = self.spl.project
        self.customer = self.project.customer
        self.service = self.spl.service

    def test_consumed_is_recalculated_properly_for_resource(self):
        calculation_time = datetime.datetime(2016, 8, 8, 15, 0)
        with freeze_time(calculation_time):
            tasks.recalculate_estimate()
            price_estimate = models.PriceEstimate.objects.get_current(scope=self.resource)

        working_minutes = (calculation_time - self.start_time).total_seconds() / 60
        expected = working_minutes * self.price_list_item.minute_rate * self.resource.disk
        self.assertAlmostEqual(price_estimate.consumed, expected)

    def test_consumed_is_recalculated_properly_for_ancestors(self):
        with freeze_time(self.start_time):
            self.second_resource = structure_factories.TestNewInstanceFactory(
                disk=10 * 1024, service_project_link=self.spl)

        calculation_time = datetime.datetime(2016, 8, 8, 15, 0)
        with freeze_time(calculation_time):
            tasks.recalculate_estimate()
            price_estimates = [models.PriceEstimate.objects.get_current(scope=ancestor) for ancestor in
                               (self.customer, self.service, self.spl, self.project)]

        working_minutes = (calculation_time - self.start_time).total_seconds() / 60
        # each ancestor is connected with 2 resources
        expected = working_minutes * self.price_list_item.minute_rate * (self.resource.disk + self.second_resource.disk)
        for price_estimate in price_estimates:
            message = 'Price estimate "consumed" is calculated wrongly for "%s". Real value: %s, expected: %s.' % (
                price_estimate.scope, price_estimate.consumed, expected)
            self.assertAlmostEqual(price_estimate.consumed, expected, msg=message)

    def test_new_estimates_are_created_in_new_month(self):
        month_start = datetime.datetime(2016, 9, 1, 0, 0)
        month_end = datetime.datetime(2016, 9, 30, 23, 59, 59)
        calculation_time = datetime.datetime(2016, 9, 1, 1, 0)
        with freeze_time(calculation_time):
            tasks.recalculate_estimate()
            price_estimates = [models.PriceEstimate.objects.get_current(scope=scope) for scope in
                               (self.resource, self.service, self.spl, self.project, self.customer)]

        total_working_minutes = int((month_end - month_start).total_seconds() / 60)
        expected_total = total_working_minutes * self.price_list_item.minute_rate * self.resource.disk
        for price_estimate in price_estimates:
            message = 'Price estimate "total" is calculated wrongly for "%s". Real value: %s, expected: %s.' % (
                price_estimate.scope, price_estimate.total, expected_total)
            self.assertAlmostEqual(price_estimate.total, expected_total, msg=message)

        working_minutes = int((calculation_time - month_start).total_seconds() / 60)
        expected_consumed = working_minutes * self.price_list_item.minute_rate * self.resource.disk
        for price_estimate in price_estimates:
            message = 'Price estimate "consumed" is calculated wrongly for "%s". Real value: %s, expected: %s.' % (
                price_estimate.scope, price_estimate.consumed, expected_consumed)
            self.assertAlmostEqual(price_estimate.consumed, expected_consumed, msg=message)



from django.contrib.contenttypes.models import ContentType
from rest_framework import test, status

from waldur_core.logging.models import Alert
from waldur_core.logging.tasks import check_threshold
from waldur_core.quotas.tests.factories import QuotaFactory
from waldur_core.structure.tests.factories import ProjectFactory, UserFactory


class QuotaThresholdAlertTest(test.APITestCase):
    def setUp(self):
        self.project = ProjectFactory()
        self.quota = self.project.quotas.get(name='nc_resource_count')

    def test_if_quota_usage_is_over_threshold_alert_is_created(self):
        self.quota.threshold = 100
        self.quota.usage = 200
        self.quota.save()

        check_threshold()

        self.assertTrue(Alert.objects.filter(
            content_type=ContentType.objects.get_for_model(self.project),
            object_id=self.project.id,
            alert_type='threshold_exceeded').exists())

    def test_if_quota_usage_is_above_threshold_alert_is_not_created(self):
        self.quota.threshold = 100
        self.quota.usage = 20
        self.quota.save()

        check_threshold()

        self.assertFalse(Alert.objects.filter(
            content_type=ContentType.objects.get_for_model(self.project),
            object_id=self.project.id,
            alert_type='threshold_exceeded').exists())

    def test_user_can_update_threshold(self):
        self.client.force_authenticate(UserFactory(is_staff=True))

        response = self.client.put(QuotaFactory.get_url(self.quota), {
            'threshold': 1000,
        })
        self.assertEqual(status.HTTP_200_OK, response.status_code, response.data)
        self.assertEqual(1000, response.data['threshold'], response.data)



from __future__ import unicode_literals

from rest_framework import test, status

from waldur_core.core import models as core_models
from waldur_core.structure.tests import factories


class BaseSshKeyTest(test.APITransactionTestCase):

    def setUp(self):
        self.staff = factories.UserFactory(is_staff=True)
        self.user = factories.UserFactory()
        self.user_key = factories.SshPublicKeyFactory(user=self.user)


class SshKeyRetrieveListTest(BaseSshKeyTest):

    def test_staff_can_get_any_key(self):
        self.client.force_authenticate(self.staff)
        url = factories.SshPublicKeyFactory.get_url(self.user_key)
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_user_can_get_own_key(self):
        self.client.force_authenticate(self.user)
        url = factories.SshPublicKeyFactory.get_url(self.user_key)
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_user_can_not_get_other_key(self):
        self.client.force_authenticate(self.user)
        shared_key = factories.SshPublicKeyFactory(user=self.staff)
        url = factories.SshPublicKeyFactory.get_url(shared_key)
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_any_user_can_get_shared_key(self):
        self.client.force_authenticate(self.user)
        shared_key = factories.SshPublicKeyFactory(user=self.staff, is_shared=True)
        url = factories.SshPublicKeyFactory.get_url(shared_key)
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)


class SshKeyCreateTest(BaseSshKeyTest):

    def test_key_user_and_name_uniqueness(self):
        self.client.force_authenticate(self.user)
        key = factories.SshPublicKeyFactory.build()
        data = {
            'name': self.user_key.name,
            'public_key': key.public_key,
        }

        response = self.client.post(factories.SshPublicKeyFactory.get_list_url(), data=data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictContainsSubset(
            {'name': ['This field must be unique.']}, response.data)

    def test_valid_key_creation(self):
        self.client.force_authenticate(self.user)
        key = factories.SshPublicKeyFactory.build()
        data = {
            'name': key.name,
            'public_key': key.public_key,
        }
        response = self.client.post(factories.SshPublicKeyFactory.get_list_url(), data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertTrue(core_models.SshPublicKey.objects.filter(name=data['name']).exists(),
                        'New key should have been created in the database')

    def test_user_cannot_add_ssh_key_with_duplicate_fingerprint(self):
        staff = factories.UserFactory(is_staff=True)
        key = factories.SshPublicKeyFactory()
        data = {
            'name': 'test',
            'public_key': key.public_key,
        }

        self.client.force_authenticate(staff)
        response = self.client.post(factories.SshPublicKeyFactory.get_list_url(), data=data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertFalse(core_models.SshPublicKey.objects.filter(**data).exists())

    def test_user_cannot_add_ssh_key_with_new_lines(self):
        staff = factories.UserFactory(is_staff=True)
        key = factories.SshPublicKeyFactory.build()
        data = {
            'name': 'test',
            'public_key': key.public_key + '\nABCD',
        }

        self.client.force_authenticate(staff)
        response = self.client.post(factories.SshPublicKeyFactory.get_list_url(), data=data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertFalse(core_models.SshPublicKey.objects.filter(**data).exists())


class SshKeyDeleteTest(BaseSshKeyTest):

    def test_staff_user_can_delete_any_key(self):
        self.client.force_authenticate(self.staff)
        response = self.client.delete(factories.SshPublicKeyFactory.get_url(self.user_key))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_user_can_delete_his_key(self):
        self.client.force_authenticate(self.user)
        response = self.client.delete(factories.SshPublicKeyFactory.get_url(self.user_key))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_user_cannot_delete_other_users_key(self):
        other_key = factories.SshPublicKeyFactory()
        self.client.force_authenticate(self.user)
        response = self.client.delete(factories.SshPublicKeyFactory.get_url(other_key))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_user_cannot_delete_shared_key(self):
        shared_key = factories.SshPublicKeyFactory(user=self.staff, is_shared=True)
        self.client.force_authenticate(self.user)

        response = self.client.delete(factories.SshPublicKeyFactory.get_url(shared_key))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
